#!/usr/bin/env python3

"""

ASE Atoms convert to LAMMPS configuration
Some functions are adapted from ASE lammpsrun.py

"""

import numpy as np
from numpy.linalg import norm

import ase.io


def dir2car(v, A):
    """Direct to cartesian coordinates"""
    return np.dot(v, A)


def car2dir(v, Ainv):
    """Cartesian to direct coordinates"""
    return np.dot(v, Ainv)


def stress9_to_stress6(s9):
    # S6: xx yy zz yz xz xy
    s6 = np.zeros(6)
    s6[0] = s9[0][0]
    s6[1] = s9[1][1]
    s6[2] = s9[2][2]
    s6[3] = s9[1][2]
    s6[4] = s9[0][2]
    s6[5] = s9[0][1]
    return s6


def stress6_to_stress9(s6):
    # S6: xx yy zz yz xz xy
    s9 = np.zeros((3, 3))
    s9[0, :] = np.array([s6[0], s6[5], s6[4]])
    s9[1, :] = np.array([s6[5], s6[1], s6[3]])
    s9[2, :] = np.array([s6[4], s6[3], s6[2]])
    return s9


def is_upper_triangular(mat):
    """
    test if 3x3 matrix is upper triangular
    LAMMPS has a rule for cell matrix definition
    """
    def near0(x):
        """Test if a float is within .00001 of 0"""
        return abs(x) < 0.00001
    return near0(mat[1, 0]) and near0(mat[2, 0]) and near0(mat[2, 1])


def convert_cell(ase_cell):
    """
    Convert a parallel piped (forming right hand basis)
    to lower triangular matrix LAMMPS can accept. This
    function transposes cell matrix so the bases are column vectors
    """

    # if ase_cell is lower triangular, cell is upper tri-angular
    cell = np.matrix.transpose(ase_cell) 

    if not is_upper_triangular(cell):
        # rotate bases into triangular matrix
        tri_mat = np.zeros((3, 3))
        A = cell[:, 0]
        B = cell[:, 1]
        C = cell[:, 2]
        tri_mat[0, 0] = norm(A)
        Ahat = A / norm(A)
        AxBhat = np.cross(A, B) / norm(np.cross(A, B))
        tri_mat[0, 1] = np.dot(B, Ahat)
        tri_mat[1, 1] = norm(np.cross(Ahat, B))
        tri_mat[0, 2] = np.dot(C, Ahat)
        tri_mat[1, 2] = np.dot(C, np.cross(AxBhat, Ahat))
        tri_mat[2, 2] = norm(np.dot(C, AxBhat))

        # create and save the transformation for coordinates
        volume = np.linalg.det(ase_cell)
        trans = np.array([np.cross(B, C), np.cross(C, A), np.cross(A, B)])
        trans = trans / volume
        coord_transform = tri_mat * trans
        return tri_mat.T # return the lower-tri-angular
    else:
        return ase_cell


def convert_positions(pos0, cell0, cell_new, direct=False):
    if direct:
        pos = np.dot(pos0, cell_new)  # positions in new cellmatrix
    else:
        cell0_inv = np.linalg.inv(cell0)
        R = np.dot(cell_new, cell0_inv)
        pos = np.dot(pos0, R)
        '''
        print(R)
        print(R.T)
        print(np.linalg.inv(R))
        print(np.linalg.det(R))
        '''
    return pos


def convert_forces(forces0, cell0, cell_new):
    forces = convert_positions(forces0, cell0, cell_new, direct=False)
    return forces


def convert_stress(s6_0, cell0, cell_new):
    s9_0 = stress6_to_stress9(s6_0)

    cell0_inv = np.linalg.inv(cell0)
    R = np.dot(cell_new, cell0_inv)
    R_T = R.T

    s9 = np.dot(np.dot(R, s9_0), R_T)
    s6 = stress9_to_stress6(s9)
    return s6


def get_atoms_ntypes(atoms):
    atomic_numbers = atoms.numbers
    Uatomic_numbers = np.unique(atomic_numbers)  # unique atomic numbers
    ntypes = Uatomic_numbers.size
    return ntypes


def set_atoms_typeids(atoms):
    csymbols = atoms.get_chemical_symbols()
    U_csymbols = np.unique(csymbols)  # unique atomic numbers
    ntypes = U_csymbols.size
    typeids = {}
    for i in range(ntypes):
        typeids[U_csymbols[i]] = i + 1  # typeid start from 1
    return typeids


def set_atoms_typeids_with_atomic_numbers(atoms):
    # set typeids as atomic numbers, which can be robust when several
    # configuration with different atomic types were used
    # however, lammps do not allow it.
    csymbols = atoms.get_chemical_symbols()
    U_csymbols = np.unique(csymbols)  # unique atomic numbers
    ntypes = U_csymbols.size
    typeids = {}
    for i in range(ntypes):
        cs = U_csymbols[i]
        typeids[cs] = ase.data.atomic_numbers[cs]
    return typeids


def get_typeid(typeids, csymbol):
    return typeids[csymbol]


def ase2lammpsdata(atoms, typeids=None, fout='out.lmp'):
    # atoms: ase.Atoms
    # typeids: eg. {'Zr': 1, 'Nb': 2, 'Hf': 3}, should start with 1 and continuous
    # fout: output file name
    fw = open(fout, 'w')
    fw.write('# LAMMPS data written by PotGen-ASE\n')
    fw.write('\n')

    # write number of atoms
    natoms = atoms.get_number_of_atoms()
    fw.write('%d atoms\n' % natoms)
    fw.write('\n')

    # write number of types
    ntypes = get_atoms_ntypes(atoms)
    fw.write("%d atom types\n" % ntypes)
    fw.write('\n')

    # write cell information
    # transfer the cell into lammps' style
    cell0 = atoms.get_cell()
    cell = convert_cell(cell0)
    xhi = cell[0, 0]
    yhi = cell[1, 1]
    zhi = cell[2, 2]

    # low triangular matrix
    xy = cell[1, 0]
    xz = cell[2, 0]
    yz = cell[2, 1]

    fw.write("%f\t%f\t xlo xhi\n" % (0, xhi))
    fw.write("%f\t%f\t ylo yhi\n" % (0, yhi))
    fw.write("%f\t%f\t zlo zhi\n" % (0, zhi))
    fw.write("%f\t%f\t%f\t xy xz yz\n" % (xy, xz, yz))
    fw.write('\n')

    # write mases
    masses = np.unique(atoms.get_masses())
    fw.write('Masses\n')
    fw.write('\n')
    for i in range(ntypes):
        fw.write('%d\t%f\n' % (i + 1, masses[i]))
        fw.flush()
    fw.write('\n')

    # convert positions
    atoms.set_cell(cell, scale_atoms=True)
    pos = atoms.get_positions()
    '''
    pos0 = atoms.get_positions()
    pos = convert_positions(pos0, cell0, cell)   # positions in new cellmatrix
    '''
    # === Write postions ===
    fw.write('Atoms\n')
    fw.write('\n')
    symbols = atoms.get_chemical_symbols()
    if typeids is None:
        typeids = set_atoms_typeids(atoms)
    for i in range(natoms):
        cs = symbols[i]
        typeid = get_typeid(typeids, cs)  # typeid start from 1~N
        # typeid = ase.data.atomic_numbers[cs]  # typeid as their atomic
        # numbers
        fw.write('%d\t%d\t%f\t%f\t%f\n' %
                 (i + 1, typeid, pos[i][0], pos[i][1], pos[i][2]))
        fw.flush()
    fw.close()
    return

# test
if __name__ == "__main__":
    import sys
    fin = sys.argv[1]
    ATOMS = ase.io.read(fin)
    ase2lammpsdata(ATOMS)
    ase2lammpsdata(ATOMS, typeids={'Al': 1}, fout=fin + '.lmp')

    sep = '=' * 40

    pos0 = ATOMS.get_positions()
    cell0 = ATOMS.get_cell()
    print(cell0)
    print(pos0)
    print(sep)

    '''
    cell_new = np.eye(3)*4.05
    '''
    delta = 4.05 * 0.02

    '''
    cell_new = 4.05*np.array([[1, delta, 0],
                         [delta, 1, 0],
                         [0, 0, 1 / (1 - delta**2)]])
    '''
    cell_new = 4.05 * np.array([[1 + delta, 0, 0],
                                [0, 1 + delta, 0],
                                [0, 0, 1 / (1 + delta**2)]])

    pos = convert_positions(pos0, cell0, cell_new)
    print(cell0)
    print(cell_new)
    #print(np.linalg.det(cell0), np.linalg.det(cell_new))
    print(pos)
    print(sep)

    # anothother transoformation for LAMMPS low-tri-angle matrix
    cell_new = convert_cell(cell_new)
    pos = convert_positions(pos0, cell0, cell_new)
    print(cell0)
    print(cell_new)
    #print(np.linalg.det(cell0), np.linalg.det(cell_new))
    print(pos)
    print(sep)

    # test for stress tensor transformation
    stress0 = np.array([0.000593	, 0.000593,	0.000593	,
                        0.,	0.00,	0.00])
    stress_new = convert_stress(stress0, cell0, cell_new)
    print(stress0)
    print(stress_new)



#!/usr/bin/env python3

import os, re, argparse, filecmp, json, glob, sys
import subprocess as sp
import numpy as np
import dpgen.auto_test.lib.vasp as vasp
import dpgen.auto_test.lib.lammps as lammps


global_equi_name = '00.equi'
global_task_name = '07.SScurve'

def cmpt_vasp(jdata, conf_dir) :
    fp_params = jdata['vasp_params']
    kspacing = fp_params['kspacing']
    kgamma = fp_params['kgamma']
    strain_direct=jdata['strain_direct']
    a,b=strain_direct[0],strain_direct[1]    

    conf_path = os.path.abspath(conf_dir)
    conf_poscar = os.path.join(conf_path, 'POSCAR')
    task_path = re.sub('confs', global_task_name, conf_path)
    task_path = os.path.join(task_path, 'vasp-k%.2f' % kspacing)

    lst_dfm_path = glob.glob(os.path.join(task_path, 'dfm-*'))
    lst_strain = []
    lst_stress = []
    print('conf_dir:',conf_dir)
    print('strain\t stress(kB)')
    for ii in lst_dfm_path :
        strain = np.loadtxt(os.path.join(ii, 'strain.out'))
        stress = vasp.get_stress(os.path.join(ii, 'OUTCAR'))
        # convert from pressure in kB to stress
        stress = -stress
        lst_strain.append(strain[a,b])
        lst_stress.append(stress[a,b])
    index=np.argsort(lst_strain)
    for ii in range(len(lst_strain)):
        print('%7.4f %7.4f'%(lst_strain[index[ii]],lst_stress[index[ii]]))
        

def cmpt_lammps(jdata, conf_dir, task_name) :
    strain_direct=jdata['strain_direct']
    a,b=strain_direct[0],strain_direct[1]

    conf_path = os.path.abspath(conf_dir)
    conf_poscar = os.path.join(conf_path, 'POSCAR')
    task_path = re.sub('confs', global_task_name, conf_path)
    task_path = os.path.join(task_path, task_name)
    

    lst_dfm_path = glob.glob(os.path.join(task_path, 'dfm-*'))
    lst_strain = []
    lst_stress = []
    print('conf_dir:',conf_dir)
    print('strain\t stress(kB)')
    for ii in lst_dfm_path :
        strain = np.loadtxt(os.path.join(ii, 'strain.out'))
        stress = lammps.get_stress(os.path.join(ii, 'log.lammps'))
        # convert from pressure to stress
        stress = -stress/1000
        lst_strain.append(strain[a,b])
        lst_stress.append(stress[a,b])
    index=np.argsort(lst_strain)
    for ii in range(len(lst_strain)):
        print('%7.4f %7.4f'%(lst_strain[index[ii]],lst_stress[index[ii]]))

def _main() :
    parser = argparse.ArgumentParser(
        description="cmpt 07.SScurve")
    parser.add_argument('TASK', type=str,
                        help='the task of generation, vasp or lammps')
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    parser.add_argument('CONF', type=str,
                        help='the path to conf')
    args = parser.parse_args()

    with open (args.PARAM, 'r') as fp :
        jdata = json.load (fp)

    print('# generate %s task with conf %s' % (args.TASK, args.CONF))
    if args.TASK == 'vasp':
        cmpt_vasp(jdata, args.CONF)               
    elif args.TASK == 'deepmd' :
        cmpt_lammps(jdata, args.CONF, args.TASK)
    elif args.TASK == 'meam' :
        cmpt_lammps(jdata, args.CONF, args.TASK)
    else :
        raise RuntimeError("unknow task ", args.TASK)
    
if __name__ == '__main__' :
    _main()



#!/usr/bin/env python3

import os,sys,json,glob,argparse,shutil
import numpy as np
import subprocess as sp
sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
from dpgen.generator.lib.pwscf import make_pwscf_input
from dpgen.generator.lib.siesta import make_siesta_input
from dpgen.generator.run import make_vasp_incar
import dpdata

def get_lmp_info(input_file) :
    with open(input_file) as fp:
        lines = [line.rstrip('\n') for line in fp]
    for ii in lines :
        words = ii.split()
        if len(words) >= 4 and words[0] == 'variable' :
            if words[1] == 'TEMP' :
                temp = float(words[3])
            elif words[1] == 'PRES' :
                pres = float(words[3])
        if len(words) >= 4 and words[0] == 'fix' :
            if words[3] == 'nvt' :
                ens = 'nvt'
            elif words[3] == 'npt' :
                ens = 'npt'
    return ens, temp, pres


def link_pp_files(tdir, fp_pp_path, fp_pp_files) :
    cwd = os.getcwd()
    os.chdir(tdir)
    for ii in fp_pp_files :
        if os.path.lexists(ii) :
            os.remove(ii)
        os.symlink(os.path.join(fp_pp_path, ii), ii)    
    os.chdir(cwd)


def copy_pp_files(tdir, fp_pp_path, fp_pp_files) :
    cwd = os.getcwd()
    os.chdir(tdir)
    for ii in fp_pp_files :
        if os.path.lexists(ii) :
            os.remove(ii)
        if os.path.exists(ii) :
            os.remove(ii)
        shutil.copyfile(os.path.join(fp_pp_path, ii), ii)    
    os.chdir(cwd)
    
        
def make_vasp(tdir, fp_params) :
    cwd = os.getcwd()
    os.chdir(tdir)
    incar = make_vasp_incar(fp_params)
    with open('INCAR', 'w') as fp:
        fp.write(incar)
    os.chdir(cwd)
        
def make_vasp_incar(tdir, fp_incar) :
    cwd = os.getcwd()
    os.chdir(tdir)
    shutil.copyfile(fp_incar, 'INCAR')
    os.chdir(cwd)        

def make_pwscf(tdir, fp_params, mass_map, fp_pp_path, fp_pp_files, user_input) : 
    cwd = os.getcwd()
    os.chdir(tdir)
    sys_data = dpdata.System('POSCAR').data
    sys_data['atom_masses'] = mass_map
    ret = make_pwscf_input(sys_data, fp_pp_files, fp_params)
    open('input', 'w').write(ret)        
    os.chdir(cwd)

def make_siesta(tdir, fp_params, fp_pp_path, fp_pp_files) :
    cwd = os.getcwd()
    os.chdir(tdir)
    sys_data = dpdata.System('POSCAR').data
    ret = make_siesta_input(sys_data, fp_pp_files, fp_params)
    open('input', 'w').write(ret)
    os.chdir(cwd)

def create_init_tasks(target_folder, param_file, output, fp_json, verbose = True) :
    target_folder = os.path.abspath(target_folder)
    output = os.path.abspath(output)
    tool_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'template')    
    jdata = json.load(open(os.path.join(target_folder, param_file)))
    fp_jdata = json.load(open(fp_json))
    # fp settings
    mass_map = jdata['mass_map']
    type_map = jdata['type_map']
    fp_style = fp_jdata['fp_style']
    fp_pp_path = fp_jdata['fp_pp_path']
    fp_pp_files = fp_jdata['fp_pp_files']
    cwd_ = os.getcwd()
    os.chdir(target_folder)
    fp_pp_path = os.path.abspath(fp_pp_path)
    os.chdir(cwd_)
    # init data sys
    init_data_prefix = jdata['init_data_prefix']
    init_data_sys = jdata['init_data_sys']
    for idx,ii in enumerate(init_data_sys):        
        sys = dpdata.LabeledSystem(os.path.join(init_data_prefix, ii), fmt = 'deepmd/npy', type_map = type_map)
        nframes = sys.get_nframes()
        sys_dir = os.path.join(output, 'init_system.%03d' % idx)
        os.makedirs(sys_dir, exist_ok = True)
        if verbose :
            print('# working on ' + sys_dir)
        with open(os.path.join(sys_dir,'record'), 'w') as fp:
            fp.write(os.path.join(init_data_prefix, ii) + '\n')
        for ff in range(nframes) :
            task_dir = os.path.join(sys_dir, 'task.%06d' % ff)
            os.makedirs(task_dir, exist_ok = True)
            sys.to_vasp_poscar(os.path.join(task_dir, 'POSCAR'), frame_idx=ff)
            # make fp
            cwd_ = os.getcwd()
            os.chdir(task_dir)
            for pp in fp_pp_files :
                if os.path.lexists(pp) :
                    os.remove(pp)
                os.symlink(os.path.relpath(os.path.join(output, pp)), pp)
            if fp_style == 'vasp':
                if os.path.lexists('INCAR') :
                    os.remove('INCAR')
                os.symlink(os.path.relpath(os.path.join(output, 'INCAR')), 'INCAR')
            elif fp_style == 'pwscf':
                try:                    
                    fp_params = fp_jdata['user_fp_params']
                    user_input = True
                except:
                    fp_params = fp_jdata['fp_params']
                    user_input = False
                make_pwscf('.', fp_params, mass_map, fp_pp_files, fp_pp_files, user_input)
            elif fp_style == 'siesta':
                make_siesta('.', fp_params, fp_pp_files, fp_pp_files)
            os.chdir(cwd_)            
    

def create_tasks(target_folder, param_file, output, fp_json, verbose = True, numb_iter = -1) :
    target_folder = os.path.abspath(target_folder)
    output = os.path.abspath(output)
    tool_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'template')    
    jdata = json.load(open(os.path.join(target_folder, param_file)))
    fp_jdata = json.load(open(fp_json))
    # goto input        
    cwd = os.getcwd()
    os.chdir(target_folder)
    sys = jdata['sys_configs']    
    # fp settings
    mass_map = jdata['mass_map']
    fp_style = fp_jdata['fp_style']
    fp_pp_path = fp_jdata['fp_pp_path']
    fp_pp_files = fp_jdata['fp_pp_files']
    cwd_ = os.getcwd()
    os.chdir(target_folder)
    fp_pp_path = os.path.abspath(fp_pp_path)
    os.chdir(cwd_)
    # fp_params = fp_jdata['fp_params']
    # collect tasks from iter dirs
    sys_tasks = [[] for ii in sys]
    sys_tasks_record = [[] for ii in sys]
    sys_tasks_cc = [0 for ii in sys]
    numb_sys = len(sys)
    iters = glob.glob('iter.[0-9]*[0-9]')
    iters.sort()
    # iters = iters[:2]
    for ii in iters[:numb_iter] :
        iter_tasks = glob.glob(os.path.join(ii, '02.fp', 'task.[0-9]*[0-9].[0-9]*[0-9]'))
        iter_tasks.sort()
        if verbose :
            print('# check iter ' + ii + ' with %6d tasks' % len(iter_tasks))
        for jj in iter_tasks :
            sys_idx = int(os.path.basename(jj).split('.')[-2])
            sys_tasks[sys_idx].append(jj)
            if os.path.islink(os.path.join(jj, 'conf.lmp')):
                linked_file = os.path.realpath(os.path.join(jj, 'conf.lmp'))
            elif os.path.islink(os.path.join(jj, 'conf.dump')):
                linked_file = os.path.realpath(os.path.join(jj, 'conf.dump'))
            else:
                raise RuntimeError('cannot file linked conf file')
            linked_keys = linked_file.split('/')
            task_record = linked_keys[-5] + '.' + linked_keys[-3] + '.' + linked_keys[-1].split('.')[0]
            task_record_keys = task_record.split('.')
            ens, temp, pres = get_lmp_info(os.path.join(ii, '01.model_devi', linked_keys[-3], 'input.lammps'))
            human_record = 'iter: %s   system: %s   model_devi_task: %s   frame: %6d   fp_task: %s   ens: %s   temp: %10.2f   pres: %10.2f' \
                           % (task_record_keys[1],
                              task_record_keys[3],
                              linked_keys[-3],
                              int(task_record_keys[-1]),
                              os.path.basename(jj),
                              ens, temp, pres
                           )
            # print(human_record)
            sys_tasks_record[sys_idx].append(human_record)            
    # for ii in range(numb_sys) :
    #     for jj in range(len(sys_tasks[ii])) :
    #         print(sys_tasks[ii][jj], sys_tasks_record[ii][jj])

    # mk output
    os.makedirs(output, exist_ok = True)
    if fp_style == 'vasp':
        copy_pp_files(output, fp_pp_path, fp_pp_files)
        make_vasp_incar(fp_params, output)
    if fp_style == 'pwscf' :
        copy_pp_files(output, fp_pp_path, fp_pp_files)        
    if fp_style == 'siesta' :
        copy_pp_files(output, fp_pp_path, fp_pp_files)
    for si in range(numb_sys) :
        sys_dir = os.path.join(output, 'system.%03d' % si)
        if verbose :
            print('# working on ' + sys_dir)
        for tt,rr in zip(sys_tasks[si], sys_tasks_record[si]) :            
            # copy poscar
            source_path = os.path.join(('iter.%s/02.fp' % rr.split()[1]), rr.split()[9])
            source_file = os.path.join(source_path, 'POSCAR')
            target_path = os.path.join(sys_dir, 'task.%06d'%sys_tasks_cc[si])
            sys_tasks_cc[si] += 1
            os.makedirs(target_path, exist_ok = True)
            target_file = os.path.join(target_path, 'POSCAR')
            target_recd = os.path.join(target_path, 'record')
            if os.path.exists(target_file) :
                os.remove(target_file)
            if os.path.exists(target_recd) :
                os.remove(target_recd)
            shutil.copyfile(source_file, target_file)
            with open(target_recd, 'w') as fp:
                fp.write('\n'.join([target_folder, rr, '']))
            # make fp
            cwd_ = os.getcwd()
            os.chdir(target_path)
            for pp in fp_pp_files :
                if os.path.lexists(pp) :
                    os.remove(pp)
                os.symlink(os.path.relpath(os.path.join(output, pp)), pp)
            if fp_style == 'vasp':
                if os.path.lexists('INCAR') :
                    os.remove('INCAR')
                os.symlink(os.path.relpath(os.path.join(output, 'INCAR')), 'INCAR')
            elif fp_style == 'pwscf':
                try:                    
                    fp_params = fp_jdata['user_fp_params']
                    user_input = True
                except:
                    fp_params = fp_jdata['fp_params']
                    user_input = False
                make_pwscf('.', fp_params, mass_map, fp_pp_files, fp_pp_files, user_input)
            elif fp_style == 'siesta':
                make_siesta('.', fp_params, mass_map, fp_pp_files, fp_pp_files)
            os.chdir(cwd_)
    os.chdir(cwd)


def _main()   :
    parser = argparse.ArgumentParser(description='Create tasks for relabeling from a DP-GEN job')
    parser.add_argument("JOB_DIR", type=str, 
                        help="the directory of the DP-GEN job")
    parser.add_argument("PARAM", type=str, default = 'fp.json',
                        help="the json file defines vasp tasks")
    parser.add_argument("OUTPUT", type=str, 
                        help="the output directory of relabel tasks")
    parser.add_argument('-p',"--parameter", type=str, default = 'param.json',
                        help="the json file provides DP-GEN paramters, should be located in JOB_DIR")
    parser.add_argument('-n',"--numb-iter", type=int, default = -1,
                        help="number of iterations to relabel")
    parser.add_argument('-v',"--verbose", action = 'store_true',
                        help="being loud")
    args = parser.parse_args()
            
    create_tasks(args.JOB_DIR, args.parameter, args.OUTPUT, args.PARAM, numb_iter = args.numb_iter, verbose = args.verbose)
    create_init_tasks(args.JOB_DIR, args.parameter, args.OUTPUT, args.PARAM, verbose = args.verbose)

if __name__ == '__main__':
    _main()
    



#!/usr/bin/python3 

import numpy as np
# from lib.vasp import system_from_poscar

def _convert_dict(idict) :
    lines = []
    for key in idict.keys() :
        if type(idict[key]) == bool:
            if idict[key] :
                ws = '.TRUE.'
            else:
                ws = '.FALSE.'
        elif type(idict[key]) == str:
            ws = '\'' + idict[key] + '\''
        else :
            ws = str(idict[key])
        lines.append('%s=%s,' % (key, ws))
    return lines


def make_pwscf_01_runctrl_dict(sys_data, idict) :
    tot_natoms = sum(sys_data['atom_numbs'])
    ntypes = len(sys_data['atom_names'])
    lines = []
    lines.append('&control')
    lines += _convert_dict(idict['control'])
    lines.append('pseudo_dir=\'./\',')
    lines.append('/')
    lines.append('&system')
    lines += _convert_dict(idict['system'])
    lines.append('ibrav=0,')
    lines.append('nat=%d,' % tot_natoms)
    lines.append('ntyp=%d,' % ntypes)
    lines.append('/')
    if 'electrons' in idict :
        lines.append('&electrons')
        lines += _convert_dict(idict['electrons'])
    lines.append('/')
    lines.append('')
    return '\n'.join(lines)
    

def _make_pwscf_01_runctrl(sys_data, ecut, ediff, smearing, degauss) :
    tot_natoms = sum(sys_data['atom_numbs'])
    ntypes = len(sys_data['atom_names'])
    ret = ""
    ret += '&control\n'
    ret += "calculation='scf',\n"
    ret += "restart_mode='from_scratch',\n"
    ret += "outdir='./OUT',\n"
    ret += "tprnfor=.TRUE.,\n"
    ret += "tstress=.TRUE.,\n"
    ret += "disk_io='none',\n"
    ret += "pseudo_dir='./',\n"
    ret += "/\n"
    ret += "&system\n"
    ret += "vdw_corr='TS',\n"
    ret += "ecutwfc=%s,\n" % str(ecut)
    ret += "ts_vdw_econv_thr=%s,\n" % str(ediff)
    ret += "nosym=.TRUE.,\n"
    ret += "ibrav=0,\n"
    ret += "nat=%d,\n" % tot_natoms
    ret += "ntyp=%d,\n" % ntypes
    if degauss is not None :
        ret += 'degauss=%f,\n' % degauss
    if smearing is not None :
        ret += 'smearing=\'%s\',\n' % (smearing.lower())
    ret += "/\n"
    ret += "&electrons\n"
    ret += "conv_thr=%s,\n" % str(ediff)
    ret += "/\n"
    return ret

def _make_pwscf_02_species(sys_data, pps) :
    atom_names = (sys_data['atom_names'])
    if 'atom_masses' in sys_data:
        atom_masses = (sys_data['atom_masses'])
    else :
        atom_masses = [1 for ii in atom_names]
    ret = ""
    ret += "ATOMIC_SPECIES\n"
    ntypes = len(atom_names)
    assert(ntypes == len(atom_names))
    assert(ntypes == len(atom_masses))
    assert(ntypes == len(pps))
    for ii in range(ntypes) :
        ret += "%s %d %s\n" % (atom_names[ii], atom_masses[ii], pps[ii])
    return ret
        
def _make_pwscf_03_config(sys_data) :
    cell = sys_data['cells'][0]
    cell = np.reshape(cell, [3,3])
    coordinates = sys_data['coords'][0]
    atom_names = (sys_data['atom_names'])
    atom_numbs = (sys_data['atom_numbs'])
    ntypes = len(atom_names)
    ret = ""
    ret += "CELL_PARAMETERS { angstrom }\n"
    for ii in range(3):
        for jj in range(3):
            ret += "%f " % cell[ii][jj]
        ret += "\n"
    ret += "\n"
    ret += "ATOMIC_POSITIONS { angstrom }\n"
    cc = 0
    for ii in range(ntypes):
        for jj in range(atom_numbs[ii]):            
            ret += "%s %f %f %f\n" % (atom_names[ii],
                                      coordinates[cc][0],
                                      coordinates[cc][1],
                                      coordinates[cc][2])
            cc += 1
    return ret

def _kshift(nkpt) :
    if (nkpt//2) * 2 == nkpt :
        return 1
    else :
        return 0
            
def _make_pwscf_04_kpoints(sys_data, kspacing):
    cell = sys_data['cells'][0]
    cell = np.reshape(cell, [3,3])
    rcell = np.linalg.inv(cell)
    rcell = rcell.T
    kpoints = [(np.ceil(2 * np.pi * np.linalg.norm(ii) / kspacing).astype(int))
               for ii in rcell]
    ret = ""
    ret += "K_POINTS { automatic }\n"
    for ii in range(3) :
        ret += "%d " % kpoints[ii]
    for ii in range(3) :
        ret += "%d " % _kshift(kpoints[ii])
    ret += "\n"
    return ret

def _make_smearing(fp_params) :
    smearing = None
    degauss = None 
    if 'smearing' in fp_params :
        smearing = (fp_params['smearing']).lower()        
    if 'sigma' in fp_params :
        degauss = fp_params['sigma']
    if (smearing is not None) and (smearing.split(':')[0] == 'mp') :
        smearing = 'mp'
    if not (smearing in [None, 'gauss', 'mp', 'fd']) :
        raise RuntimeError("unknow smearing method " + smearing)
    return smearing, degauss

def make_pwscf_input(sys_data, fp_pp_files, fp_params, user_input = True) :
    if not user_input :
        ecut = fp_params['ecut']
        ediff = fp_params['ediff']
        smearing, degauss = _make_smearing(fp_params)
    kspacing = fp_params['kspacing']
    ret = ""
    if not user_input:
        ret += _make_pwscf_01_runctrl(sys_data, ecut, ediff, smearing, degauss)
    else :
        ret += make_pwscf_01_runctrl_dict(sys_data, fp_params)
    ret += "\n"
    ret += _make_pwscf_02_species(sys_data, fp_pp_files)
    ret += "\n"
    ret += _make_pwscf_03_config(sys_data)
    ret += "\n"
    ret += _make_pwscf_04_kpoints(sys_data, kspacing)
    ret += "\n"
    return ret

ry2ev = 13.605693009
bohr2ang = 0.52917721067
kbar2evperang3 = 1

def get_block (lines, keyword, skip = 0) :
    ret = []
    for idx,ii in enumerate(lines) :
        if keyword in ii :
            blk_idx = idx + 1 + skip
            while len(lines[blk_idx]) != 0 and blk_idx != len(lines):
                ret.append(lines[blk_idx])
                blk_idx += 1
            break
    return ret

def get_types (lines) :
    ret = []
    blk = get_block(lines, 'ATOMIC_SPECIES')
    for ii in blk:
        ret.append(ii.split()[0])
    return ret

def get_cell (lines) :
    ret = []
    blk = get_block(lines, 'CELL_PARAMETERS')
    for ii in blk:
        ret.append([float(jj) for jj in ii.split()[0:3]])
    ret = np.array([ret])
    return ret

def get_coords (lines) :
    ret = []
    blk = get_block(lines, 'ATOMIC_POSITIONS')
    for ii in blk:
        ret.append([float(jj) for jj in ii.split()[1:4]])
    ret = np.array([ret])
    return ret

def get_natoms (lines) :
    types = get_types (lines)
    names = []
    blk = get_block(lines, 'ATOMIC_POSITIONS')
    for ii in blk:
        names.append(ii.split()[0])
    natoms = []
    for ii in types :
        natoms.append(names.count(ii))
    # return np.array(natoms, dtype = int)
    return natoms

def get_atom_types(lines) :
    types = get_types (lines)
    names = []
    blk = get_block(lines, 'ATOMIC_POSITIONS')
    for ii in blk:
        names.append(ii.split()[0])
    ret = []
    for ii in names:
        ret.append(types.index(ii))
    return np.array(ret, dtype = int)

def get_energy (lines) :
    for ii in lines :
        if '!    total energy' in ii :
            return np.array([ry2ev * float(ii.split('=')[1].split()[0])])
    return None

def get_force (lines) :
    blk = get_block(lines, 'Forces acting on atoms', skip = 1)
    ret = []
    for ii in blk:
        ret.append([float(jj) for jj in ii.split('=')[1].split()])
    ret = np.array([ret])
    ret *= (ry2ev / bohr2ang)
    return ret

def get_stress (lines, cells) :
    vols = []
    for ii in cells:
        vols.append(np.linalg.det(ii.reshape([3,3])))
    blk = get_block(lines, 'total   stress')
    ret = []
    for ii in blk:
        ret.append([float(jj) for jj in ii.split()[3:6]])
    ret = np.array([ret])
    for idx,ii in enumerate(ret):
        ii *= vols[idx] * 1e3 / 1.602176621e6
    return ret
    

def cvt_1frame (fin, fout):
    with open(fout) as fp:
        outlines = fp.read().split('\n')
    with open(fin) as fp:
        inlines = fp.read().split('\n')
    # outlines = open(fout, 'r').read().split('\n')
    # inlines = open(fin, 'r').read().split('\n')
    data = {}
    data['orig'] = np.array([0,0,0])
    data['atom_names'] = (get_types (inlines))
    data['atom_numbs'] = (get_natoms(inlines))
    data['atom_types'] = (get_atom_types(inlines))
    data['coords'] = (get_coords(inlines))
    data['cells'] = (get_cell  (inlines))
    data['energies'] = (get_energy(outlines))
    data['forces'] = (get_force (outlines))
    data['virials'] = (get_stress(outlines, data['cells']))
    return data




#!/usr/bin/env python3

import sys
import os
import shutil
import glob
import argparse

import numpy as np
import ase.io
import io_lammps


def create_disturbs_atomsk(fin, nfile, dmax=1.0, ofmt="lmp"):
    # removing the exists files
    flist = glob.glob('*.' + ofmt)
    for f in flist:
        os.remove(f)
    # Based on our tests, we find it always creates a disturb by
    # constant value of dmax for atomsk
    for i in range(1, nfile + 1):
        fout = fin + str(i) + '.' + ofmt
        cmd = "atomsk " + fin + " -disturb " + str(dmax) + " -wrap -ow " + fout
        os.system(cmd)
    return


def random_range(a, b, ndata=1):
    data = np.random.random(ndata) * (b - a) + a
    return data


def gen_random_disturb(dmax, a, b, dstyle='uniform'):
    d0 = np.random.rand(3) * (b - a) + a
    dnorm = np.linalg.norm(d0)
    if dstyle == 'normal':
        dmax = np.random.standard_normal(0, 0.5) * dmax
    elif dstyle == 'constant':
        pass
    else:
        # use if we just wanna a disturb in a range of [0, dmax),
        dmax = np.random.random() * dmax
    dr = dmax / dnorm * d0
    return dr


def create_disturbs_ase(fin, nfile, dmax=1.0, ofmt="lmp", dstyle='uniform', write_d=False):
    # removing the exists files
    flist = glob.glob('*.' + ofmt)
    for f in flist:
        os.remove(f)

    # read-in by ase
    atoms = ase.io.read(fin)
    natoms = atoms.get_number_of_atoms()
    pos0 = atoms.get_positions()

    # creat nfile ofmt files.
    for fid in range(1, nfile + 1):
        dpos = np.zeros((natoms, 3))
        atoms_d = atoms.copy()
        if write_d:
            fw = open('disp-' + str(fid) + '.dat', 'w')
        for i in range(natoms):
            # Use copy(), otherwise it will modify the input atoms every time.
            dr = gen_random_disturb(dmax, -0.5, 0.5, dstyle)
            '''
            if i == 1:
                print(dr)
                print(np.linalg.norm(dr))
            '''
            dpos[i, :] = dr
            if write_d:
                dnorm = np.linalg.norm(dr)
                fw.write('%d\t%f\t%f\t%f\t%f\n' %
                         (i + 1, dr[0], dr[1], dr[2], dnorm))
                fw.flush()
        pos = pos0 + dpos
        atoms_d.set_positions(pos)
        fout = fin + str(fid) + '.' + ofmt
        print("Creating %s ..." % fout)
        if ofmt in ['lmp', 'lammps_data']:
            # for lammps, use my personal output functions
            io_lammps.ase2lammpsdata(atoms_d, fout)
        else:
            ase.io.write(fout, atoms_d, ofmt, vasp5 = True)
        if write_d:
            fw.close()
    return


def gen_random_emat(etmax, diag=0):
    if np.abs(etmax) >= 1e-6:
        e = random_range(-etmax, etmax, 6)
    else:
        e = np.zeros(6)
    if diag != 0:
        # isotropic behavior
        e[3], e[4], e[5] = 0, 0, 0
    emat = np.array(
        [[e[0], 0.5 * e[5], 0.5 * e[4]],
         [0.5 * e[5], e[1], 0.5 * e[3]],
         [0.5 * e[4], 0.5 * e[3], e[2]]]
    )
    emat = emat + np.eye(3)
    return emat


def create_disturbs_ase_dev(fin, nfile, dmax=1.0, etmax=0.1, ofmt="lmp", dstyle='uniform', write_d=False, diag=0):
    # removing the exists files
    flist = glob.glob('*.' + ofmt)
    for f in flist:
        os.remove(f)

    # read-in by ase
    atoms = ase.io.read(fin)
    natoms = atoms.get_number_of_atoms()
    cell0 = atoms.get_cell()

    # creat nfile ofmt files.
    for fid in range(1, nfile + 1):
        # Use copy(), otherwise it will modify the input atoms every time.
        atoms_d = atoms.copy()

        # random flux for atomic positions
        if write_d:
            fw = open('disp-' + str(fid) + '.dat', 'w')
        dpos = np.zeros((natoms, 3))
        for i in range(natoms):
            dr = gen_random_disturb(dmax, -0.5, 0.5, dstyle)
            dpos[i, :] = dr
            if write_d:
                dnorm = np.linalg.norm(dr)
                fw.write('%d\t%f\t%f\t%f\t%f\n' %
                         (i + 1, dr[0], dr[1], dr[2], dnorm))
                fw.flush()

        # random flux for volumes
        cell = np.dot(cell0, gen_random_emat(etmax, diag))
        atoms_d.set_cell(cell, scale_atoms=True)
        if write_d:
            fout_c = 'cell-' + str(fid) + '.dat'
            np.savetxt(fout_c, cell, '%f')

        # determine new cell & atomic positions randomiziations
        pos = atoms_d.get_positions() + dpos
        atoms_d.set_positions(pos)

        # pre-converting the Atoms to be in low tri-angular cell matrix
        cell_new = io_lammps.convert_cell(cell)
        #pos_new = io_lammps.convert_positions(pos, cell, cell_new)
        atoms_d.set_cell(cell_new, scale_atoms=True)
        # atoms_d.set_positions(pos_new)

        # Writing it
        fout = fin + str(fid) + '.' + ofmt
        print("Creating %s ..." % fout)
        if ofmt in ['lmp', 'lammps_data']:
            # for lammps, use my personal output functions
            io_lammps.ase2lammpsdata(atoms_d, fout=fout)
        else:
            ase.io.write(fout, atoms_d, ofmt, vasp5 = True)
        if write_d:
            fw.close()
    return


def create_random_alloys(fin, alloy_dist, ifmt='vasp', ofmt='vasp'):
    '''
    In fact, atomsk also gives us the convinient tool to do this
    '''
    # alloy_dist = {'Zr': 0.80, 'Nb': 0.20}
    atomic_symbols = alloy_dist.keys()
    atomic_ratios = alloy_dist.values()

    # renormalize the atomic_ratio
    atomic_ratios = atomic_ratios / atomic_ratios.sum()

    atoms = ase.io.read(fin, format=ifmt)

    # decide the exactly number of atoms for each types.
    natoms = atoms.num_of_numbers()
    num_for_each = []
    for r in atomic_ratios:
        num_for_each.append(int(r * natoms))

    # ========= decide which atom-ID to be substituted. ===========
    # a natoms vector with random sequence
    all_ids = np.random.permutation(natoms)
    ids_for_each = []
    for i in range(num_for_each):
        num = num_for_each[i]
        if i == 0:
            id_start = 0
        else:
            id_start = num_for_each[i - 1]
        ids_for_each.append(all_ids[id_start, id_start + num])

    # subsitute them, by replacing their chemical symbols of A by B
    new_chemical_symbols = atoms.get_chemical_symbols()
    for i in range(1, num_for_each):  # ignore the 0-th element (it is itself)
        symbol = atomic_symbols[i]
        ids = ids_for_each[i]
        for idx in ids:
            new_chemical_symbols[idx] = symbol
    # set the new chemical_symbols to the atoms
    atoms.set_chemical_symbols(new_chemical_symbols)

    # write it as ofmt
    fout = fin.split(',')[0] + '_random' + ofmt
    ase.io.write(fout, atoms, format=ofmt, vasp5 = True)
    return


def RandomDisturbParser():
    parser = argparse.ArgumentParser(
        description="Script to generate random disturb configurations")
    parser.add_argument('fin', type=str, help="input file name")
    parser.add_argument('nfile', type=int,
                        help='number of files to be created')
    parser.add_argument('dmax', type=float, help='dmax')
    parser.add_argument('-etmax', type=float, default=0,
                        help='etmax for random strain tensor generations')
    parser.add_argument('-diag', type=int, default=0,
                        help='only diagonal elements of strain tensors are randomized?')
    parser.add_argument('-ofmt', type=str, default='lmp',
                        help='output fileformat')
    parser.add_argument('-dstyle', type=str, default='uniform',
                        help='random distribution style [uniform?]')
    parser.add_argument('-wd', '--write_disp', type=int,
                        default=0, help='write displacement information?')
    return parser.parse_args()


#
if __name__ == "__main__":
    args = RandomDisturbParser()
    fin = args.fin
    nfile = args.nfile
    dmax = args.dmax
    ofmt = args.ofmt
    dstyle = args.dstyle
    write_d = args.write_disp
    etmax = args.etmax
    diag = args.diag

    if write_d == 0:
        write_d = False
    else:
        write_d = True

    # main program
    #create_disturbs_atomsk(fin, nfile, dmax, ofmt)
    #create_disturbs_ase(fin, nfile, dmax, ofmt, dstyle, write_d)
    create_disturbs_ase_dev(fin, nfile, dmax, etmax,
                            ofmt, dstyle, write_d, diag)



import os,sys,json,glob,shutil
import dpdata
import numpy as np
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'auto_test'
from .context import make_kspacing_kpoints
from .context import setUpModule
from pymatgen.io.vasp import Kpoints,Incar

class TestVASPMakeKpoint(unittest.TestCase):
    def test_gf_all (self):
        kspacing = 0.16
        gamma = False
        all_test = glob.glob(os.path.join('data.vasp.kp.gf', 'test.*'))
        for ii in all_test :            
            ret=make_kspacing_kpoints(os.path.join(ii, 'POSCAR'), kspacing, gamma)
            kp = [int(jj) for jj in (ret.split('\n')[3].split())]
            kp_ref = list(np.loadtxt(os.path.join(ii, 'kp.ref'), dtype = int))
            self.assertTrue(kp == kp_ref)




#!/usr/bin/env python
# coding: utf-8

from dpgen.dispatcher.SSHContext import SSHSession
from dpgen.dispatcher.SSHContext import SSHContext
from dpgen.dispatcher.Slurm import Slurm
from dpgen import dlog
import os
import json
import numpy as np


def decide_train_machine(mdata):
	if 'train' in mdata:
	    continue_flag = False
	    if 'record.machine' in os.listdir():
	        try:
	            with open('record.machine', 'r') as _infile:
	                profile = json.load(_infile)
	                if profile['purpose'] == 'train':
	                    mdata['train_machine'] = profile['machine']
	                    mdata['train_resources'] = profile['resources']
	                    if 'deepmd_path' in profile:
	                    	mdata['deepmd_path'] = profile['deepmd_path']
	                    elif 'python_path' in profile:
	                    	mdata['python_path'] = profile['python_path']
	                    if "group_size" in profile:
	                    	mdata["train_group_size"] = profile["group_size"]
	                    if 'deepmd_version' in profile:
	                    	mdata["deepmd_version"] = profile['deepmd_version']
	                    continue_flag = True
	        except:
	            pass
	    if ("hostname" not in mdata["train"][0]["machine"]) or (len(mdata["train"]) == 1):
	    	mdata["train_machine"] = mdata["train"][0]["machine"]
	    	mdata["train_resources"] = mdata["train"][0]["resources"]
	    	if 'deepmd_path' in mdata["train"][0]:
	    		mdata["deepmd_path"] = mdata["train"][0]["deepmd_path"]
	    	elif 'python_path' in mdata["train"][0]:
	    		mdata["python_path"] = mdata["train"][0]["python_path"]
	    	if "group_size" in mdata["train"][0]:
	    		mdata["train_group_size"] = mdata["train"][0]["group_size"]
	    	if 'deepmd_version' in mdata["train"][0]:
	    		mdata["deepmd_version"] = mdata["train"][0]["deepmd_version"]
	    	if 'command' in mdata["train"][0]:
	    		mdata["train_command"] = mdata["train"][0]["command"]
	    	continue_flag = True
		
	    pd_flag = False
	    pd_count_list =[]
	    # pd for pending job in slurm
	    # if we need to launch new machine_idxines
	    if not continue_flag:

	        #assert isinstance(mdata['train']['machine'], list)
	        #assert isinstance(mdata['train']['resources'], list)
	        #assert len(mdata['train']['machine']) == len(mdata['train']['resources'])
	        # mdata['train'] is  a list 
	        for machine_idx in range(len(mdata['train'])):
	            temp_machine = mdata['train'][machine_idx]['machine']
	            temp_resources = mdata['train'][machine_idx]['resources']
	            temp_ssh_sess = SSHSession(temp_machine)
	            cwd = os.getcwd()
	            temp_context = SSHContext(cwd, temp_ssh_sess)
	            temp_batch = Slurm(temp_context)
	            command = temp_batch._make_squeue(temp_machine, temp_resources)
	            ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
	            pd_response = stdout.read().decode('utf-8').split("\n")
	            pd_count = len(pd_response)
	            temp_context.clean()
	            ## If there is no need to waiting for allocation
	            if pd_count ==1:
	                mdata['train_machine'] = temp_machine   
	                mdata['train_resources'] = temp_resources
	                if 'deepmd_path' in mdata['train'][machine_idx]:
	                	mdata['deepmd_path'] = mdata['train'][machine_idx]['deepmd_path']
	                elif 'python_path' in mdata['train'][machine_idx]:
	                	mdata['python_path'] = mdata['train'][machine_idx]['python_path']
	                if 'group_size' in mdata['train'][machine_idx]:
	                	mdata['train_group_size'] = mdata['train'][machine_idx]['group_size']
	                if 'deepmd_version' in mdata['train'][machine_idx]:
	                	mdata['deepmd_version'] = mdata['train'][machine_idx]['deepmd_version']

	                ## No need to wait
	                pd_flag = True
	                break
	            else:
	                pd_count_list.append(pd_count)
	        if not pd_flag:
	        	## All machines need waiting, then compare waiting jobs
	        	## Select a machine which has fewest waiting jobs
	            min_machine_idx = np.argsort(pd_count_list)[0]
	            mdata['train_machine'] = mdata['train'][min_machine_idx]['machine']
	            mdata['train_resources'] = mdata['train'][min_machine_idx]['resources']
	            if 'deepmd_path' in mdata['train'][min_machine_idx]:
	            	mdata['deepmd_path'] = mdata['train'][min_machine_idx]['deepmd_path']
	            elif 'python_path' in mdata['train'][min_machine_idx]:
	            	mdata['python_path'] = mdata['train'][min_machine_idx]['python_path']
	            if "group_size" in mdata['train'][min_machine_idx]:
	            	mdata["train_group_size"] = mdata['train'][min_machine_idx]["group_size"]
	            if 'deepmd_version' in mdata['train'][min_machine_idx]:
	            	mdata['deepmd_version'] = mdata['train'][min_machine_idx]["deepmd_version"]

            ## Record which machine is selected
	        with open("record.machine","w") as _outfile:
	            profile = {}
	            profile['purpose'] = 'train'
	            profile['machine'] = mdata['train_machine']
	            profile['resources'] = mdata['train_resources']
	            if 'deepmd_path' in mdata:
	            	profile['deepmd_path'] = mdata['deepmd_path']
	            elif 'python_path' in mdata:
	            	profile['python_path'] = mdata['python_path']
	            if "train_group_size" in mdata:
	            	profile["group_size"] = mdata["train_group_size"]
	            if 'deepmd_version' in mdata:
	            	profile['deepmd_version'] = mdata['deepmd_version']

	            json.dump(profile, _outfile, indent = 4)
	return mdata

def decide_model_devi_machine(mdata):
	if 'model_devi' in mdata:
	    continue_flag = False
	    if 'record.machine' in os.listdir():
	        try:
	            with open('record.machine', 'r') as _infile:
	                profile = json.load(_infile)
	                if profile['purpose'] == 'model_devi':
	                    mdata['model_devi_machine'] = profile['machine']
	                    mdata['model_devi_resources'] = profile['resources']
	                    mdata['lmp_command'] = profile['command']
	                    mdata['model_devi_group_size'] = profile['group_size']
	                    continue_flag = True
	        except:
	            pass
	    if ("hostname" not in mdata["model_devi"][0]["machine"]) or (len(mdata["model_devi"]) == 1):
	    	mdata["model_devi_machine"] = mdata["model_devi"][0]["machine"]
	    	mdata["model_devi_resources"] = mdata["model_devi"][0]["resources"]
	    	mdata["lmp_command"] = mdata["model_devi"][0]["command"]
	    	#if "group_size" in mdata["train"][0]:
	    	mdata["model_devi_group_size"] = mdata["model_devi"][0]["group_size"]
	    	continue_flag = True

	    pd_count_list =[]
	    pd_flag = False
	    if not continue_flag:

	        #assert isinstance(mdata['model_devi']['machine'], list)
	        #ssert isinstance(mdata['model_devi']['resources'], list)
	        #assert len(mdata['model_devi']['machine']) == len(mdata['model_devi']['resources'])
	    
	        for machine_idx in range(len(mdata['model_devi'])):
	            temp_machine = mdata['model_devi'][machine_idx]['machine']
	            temp_resources = mdata['model_devi'][machine_idx]['resources']
	            #assert isinstance(temp_machine, dict), "unsupported type of model_devi machine [%d]!" %machine_idx
	            #assert isinstance(temp_resources, dict), "unsupported type of model_devi resources [%d]!"%machine_idx
	            #assert temp_machine['machine_type'] == 'slurm', "Currently only support for Slurm!"
	            temp_ssh_sess = SSHSession(temp_machine)
	            cwd = os.getcwd()
	            temp_context = SSHContext(cwd, temp_ssh_sess)
	            temp_batch = Slurm(temp_context)
	            command = temp_batch._make_squeue(temp_machine, temp_resources)
	            ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
	            pd_response = stdout.read().decode('utf-8').split("\n")
	            pd_count = len(pd_response)
	            temp_context.clean()
	            if pd_count ==0:
	                mdata['model_devi_machine'] = temp_machine   
	                mdata['model_devi_resources'] = temp_resources
	                mdata['lmp_command'] = mdata['model_devi'][machine_idx]['command']
	                mdata['model_devi_group_size'] =  mdata['model_devi'][machine_idx]['group_size']
	                pd_flag = True
	                break
	            else:
	                pd_count_list.append(pd_count)
	        if not pd_flag:
	            min_machine_idx = np.argsort(pd_count_list)[0]
	            mdata['model_devi_machine'] = mdata['model_devi'][min_machine_idx]['machine']
	            mdata['model_devi_resources'] = mdata['model_devi'][min_machine_idx]['resources']
	            mdata['lmp_command'] = mdata['model_devi'][min_machine_idx]['command']
	            mdata['model_devi_group_size'] =  mdata['model_devi'][min_machine_idx]['group_size']
	        with open("record.machine","w") as _outfile:
	            profile = {}
	            profile['purpose'] = 'model_devi'
	            profile['machine'] = mdata['model_devi_machine']
	            profile['resources'] = mdata['model_devi_resources']
	            profile['group_size'] = mdata['model_devi_group_size']
	            profile['command'] = mdata['lmp_command']

	            json.dump(profile, _outfile, indent = 4)
	return mdata
def decide_fp_machine(mdata):

	if 'fp' in mdata:
	    #ssert isinstance(mdata['fp']['machine'], list)
	    #assert isinstance(mdata['fp']['resources'], list)
	    #assert len(mdata['fp']['machine']) == len(mdata['fp']['resources'])
	    continue_flag = False
	    ## decide whether to use an existing machine
	    if 'record.machine' in os.listdir():
	        try:
	            with open('record.machine', 'r') as _infile:
	                profile = json.load(_infile)
	                if profile['purpose'] == 'fp':
	                    mdata['fp_machine'] = profile['machine']
	                    mdata['fp_resources'] = profile['resources']
	                    mdata['fp_command'] = profile['command']
	                    mdata['fp_group_size'] = profile['group_size']
	                    #mdata['deepmd_path'] = profile['deepmd_path']
	                    continue_flag = True
	        except:
	            pass
	    if ("hostname" not in mdata["fp"][0]["machine"]) or (len(mdata["fp"]) == 1):
	    	mdata["fp_machine"] = mdata["fp"][0]["machine"]
	    	mdata["fp_resources"] = mdata["fp"][0]["resources"]
	    	mdata["fp_command"] = mdata["fp"][0]["command"]
	    	#if "group_size" in mdata["train"][0]:
	    	mdata["fp_group_size"] = mdata["fp"][0]["group_size"]
	    	continue_flag = True

	   
	    pd_count_list =[]
	    pd_flag = False
	    if not continue_flag:
		    for machine_idx in range(len(mdata['fp'])):
		        temp_machine = mdata['fp'][machine_idx]['machine']
		        temp_resources = mdata['fp'][machine_idx]['resources']
		        temp_ssh_sess = SSHSession(temp_machine)
		        cwd = os.getcwd()
		        temp_context = SSHContext(cwd, temp_ssh_sess)
		        temp_batch = Slurm(temp_context)
		        command = temp_batch._make_squeue(temp_machine, temp_resources)
		        ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
		        pd_response = stdout.read().decode('utf-8').split("\n")
		        pd_count = len(pd_response)
		        temp_context.clean()
		        #dlog.info(temp_machine["username"] + " " + temp_machine["hostname"] +  " " + str(pd_count))
		        if pd_count ==0:
		            mdata['fp_machine'] = temp_machine   
		            mdata['fp_resources'] = temp_resources
		            mdata['fp_command'] = mdata['fp'][machine_idx]['command']
		            mdata['fp_group_size'] =  mdata['fp'][machine_idx]['group_size']
		            pd_flag = True
		            break
		        else:
		            pd_count_list.append(pd_count)
		    if not pd_flag:
		        min_machine_idx = np.argsort(pd_count_list)[0]
		        mdata['fp_machine'] = mdata['fp'][min_machine_idx]['machine']
		        mdata['fp_resources'] = mdata['fp'][min_machine_idx]['resources']
		        mdata['fp_command'] = mdata['fp'][min_machine_idx]['command']
		        mdata['fp_group_size'] =  mdata['fp'][min_machine_idx]['group_size']

		    with open("record.machine","w") as _outfile:
		            profile = {}
		            profile['purpose'] = 'fp'
		            profile['machine'] = mdata['fp_machine']
		            profile['resources'] = mdata['fp_resources']
		            profile['group_size'] = mdata['fp_group_size']
		            profile['command'] = mdata['fp_command']
		            json.dump(profile, _outfile, indent = 4)
	return mdata




#!/usr/bin/env python3

from __future__ import division
import os
import sys
import argparse

import numpy as np
from scipy.optimize import leastsq, root, fsolve, curve_fit
from scipy.optimize import minimize
from scipy.misc import derivative
from scipy.interpolate import *
import scipy.integrate as INT
import matplotlib.pyplot as plt

kb = 1.3806488e-23  # J K^-1
kb_ev = 8.6173324e-05  # eV K^-1
h = 6.62606957e-34      # J.s
h_ev = 4.135667516e-15  # eV s
hb = 1.054571726e-34  # J.s
hb_ev = 6.58211928e-16  # eV s
mu = 1.660538921e-27  # kg
me = 9.1093821545e-31  # kg
NA = 6.02214129e+23  # number per mol

eV2GPa = 1.602176565e+2
eV2mol = 9.648455461e4  # eV/K --> J/mol/K


def __version__():
    return '1.2.5'


def get_eos_list_4p():
    eos_list_4p = [
        'murnaghan', 'birch',
        'BM4', 'mBM4', 'mBM4poly', 'rBM4',
        'rPT4', 'LOG4',
        'vinet', 'Li4p', 'universal',
        'morse', 'morse_AB', 'mie', 'mie_simple',
        'SJX_v2'
    ]
    return eos_list_4p


def get_eos_list_5p():
    eos_list_5p = [
        'BM5', 'mBM5', 'rBM5', 'mBM5poly',
        'rPT5', 'LOG5',
        'TEOS', 'SJX_5p']
    return eos_list_5p


def get_eos_list_6p():
    eos_list_6p = ['morse_6p']
    return eos_list_6p


def get_eos_list_3p():
    eos_list_3p = ['morse_3p']
    return eos_list_3p


def get_eos_list():
    LIST_ALL = get_eos_list_3p() + get_eos_list_4p() + \
        get_eos_list_5p() + get_eos_list_6p()
    return LIST_ALL


# ----------------------------------------------------------------------------------------
def res_murnaghan(pars, y, x):
    return y - murnaghan(x, pars)


def murnaghan(vol, pars):
    """
    Four-parameters murnaghan EOS.
    From PRB 28,5480 (1983)
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    xx = (v0 / vol)**bp
    A = e0 - b0 * v0 / (bp - 1)
    B = b0 * vol / bp
    ee = A + B * (1 + xx / (bp - 1))

    # ee = e0+b0*vol/bp*(((v0/vol)**bp)/(bp-1)+1) -b0*v0/(bp-1)

    return ee

# ----------------------------------------------------------------------------------------


def res_birch(pars, y, x):
    return y - birch(x, pars)


def birch(v, parameters):
    """
    From Intermetallic compounds: Principles and Practice, Vol. I: Princples
    Chapter 9 pages 195-210 by M. Mehl. B. Klein, D. Papaconstantopoulos
    paper downloaded from Web

    case where n=0
    """
    e0 = parameters[0]
    b0 = parameters[1]
    bp = parameters[2]
    v0 = parameters[3]

    e = (e0 + 9.0 / 8.0 * b0 * v0 * ((v0 / v) ** (2.0 / 3.0) - 1.0) ** 2
         + 9.0 / 16.0 * b0 * v0 * (bp - 4.) * ((v0 / v) ** (2.0 / 3.0) - 1.0) ** 3)

    return e


# ----------------------------------------------------------------------------------------
def res_mBM4(pars, y, x):
    return y - mBM4(x, pars)


def calc_props_mBM4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (-74 + 45 * bp - 9 * bp * bp) / (9 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def mBM4(vol, pars):
    """
    Birch-Murnaghan 4 pars equation from PRB 70, 224107, 3-order BM
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    a = e0 + 9 * b0 * v0 * (4 - bp) / 2.
    b = -9 * b0 * v0 ** (4. / 3) * (11 - 3 * bp) / 2.
    c = 9 * b0 * v0 ** (5. / 3) * (10 - 3 * bp) / 2.
    d = -9 * b0 * v0 ** 2 * (3 - bp) / 2.

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -n / 3)
    ee = a + b * VV + c * VV ** 2 + d * VV ** 3

    return ee


def res_mBM5(pars, y, x):
    return y - mBM5(x, pars)


def mBM5(vol, pars):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    b2p = pars[4]

    '''
    # copy from ShunLi's matlab scripts.
    a = (8 * e0 + 3 * b0 * (122 + 9 * b0 * b2p - 57 * bp + 9 * bp * bp) * v0) / 8
    b = (-3 * b0 * (107 + 9 * b0 * b2p - 54 * bp + 9 * bp * bp) * v0 ** (4 / 3)) / 2
    c = (9 * b0 * (94 + 9 * b0 * b2p - 51 * bp + 9 * bp * bp) * v0 ** (5 / 3)) / 4
    d = (-3 * b0 * (83 + 9 * b0 * b2p - 48 * bp + 9 * bp * bp) * v0 ** 2) / 2
    e = (3 * b0 * (74 + 9 * b0 * b2p - 45 * bp + 9 * bp * bp) * v0 ** (7 / 3)) / 8
    '''
    # retype according to formula in the article
    a = e0 + 3 * b0 * v0 * (122 + 9 * b0 * b2p - 57 * bp + 9 * bp * bp) / 8
    b = -3 * b0 * v0**(4 / 3) * (107 + 9 * b0 *
                                 b2p - 54 * bp + 9 * bp * bp) / 2
    c = 9 * b0 * v0**(5 / 3) * (94 + 9 * b0 * b2p - 51 * bp + 9 * bp * bp) / 4
    d = -3 * b0 * v0**2 * (83 + 9 * b0 * b2p - 48 * bp + 9 * bp * bp) / 2
    e = 3 * b0 * v0**(7 / 3) * (74 + 9 * b0 * b2p - 45 * bp + 9 * bp * bp) / 8

    VV = np.power(vol, -1 / 3)
    ee = a + b * VV + c * VV ** 2 + d * VV ** 3 + e * VV ** 4

    return ee


# ----------------------------------------------------------------------------------------
def res_mBM4poly(pars, y, x):
    return y - mBM4poly(x, pars)


def mBM4poly(vol, parameters):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048, original expressions.
    """
    a = parameters[0]
    b = parameters[1]
    c = parameters[2]
    d = parameters[3]
    e = 0

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -1 / 3)
    E = a + b * VV + c * VV ** 2 + d * VV ** 3 + e * VV ** 4
    return E


def calc_v0_mBM4poly(x, pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = 0

    f = ((4 * e) / (3 * x ** (7 / 3)) + d / x ** 2 + (2 * c) /
         (3 * x ** (5 / 3)) + b / (3 * x ** (4 / 3))) * eV2GPa
    return f


def calc_props_mBM4poly(pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = 0

    v0 = 4 * c ** 3 - 9 * b * c * d + \
        np.sqrt((c ** 2 - 3 * b * d) * (4 * c ** 2 - 3 * b * d) ** 2)
    v0 = -v0 / b ** 3
    b0 = ((28 * e) / (9 * v0 ** (10 / 3)) +
          (2 * d) / v0 ** 3 + (10 * c) / (9 * v0 ** (8 / 3)) +
          (4 * b) / (9 * v0 ** (7 / 3))) * v0
    bp = (98 * e + 54 * d * v0 ** (1 / 3) + 25 * c * v0 ** (2 / 3) + 8 * b * v0) / \
        (42 * e + 27 * d * v0 ** (1 / 3) + 15 * c * v0 ** (2 / 3) + 6 * b * v0)
    b2p = (v0**(8 / 3) * (9 * d * (14 * e + 5 * c * v0**(2 / 3) + 8 * b * v0) +
                          2 * v0**(1 / 3) * (126 * b * e * v0 ** (1 / 3) + 5 * c * (28 * e + b * v0)))) / \
        (2 * (14 * e + 9 * d * v0**(1 / 3) + 5 * c * v0**(2 / 3) + 2 * b * v0)**3)
    e0 = mBM4poly(v0, pars)

    props = [e0, b0, bp, v0, b2p]
    return props

#---------------------------------


def res_mBM5poly(pars, y, x):
    return y - mBM5poly(x, pars)


def mBM5poly(vol, pars):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048, original expressions.
    """
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -1 / 3)
    E = a + b * VV + c * VV ** 2 + d * VV ** 3 + e * VV ** 4
    return E


def calc_v0_mBM5poly(x, pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]

    f = ((4 * e) / (3 * x ** (7 / 3)) + d / x ** 2 + (2 * c) /
         (3 * x ** (5 / 3)) + b / (3 * x ** (4 / 3))) * eV2GPa
    return f


def calc_props_mBM5poly(pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]
    n = 1

    # guess v0
    vtest = np.linspace(1, 100, 101)
    etest = mBM5poly(vtest, pars)
    vg = vtest[etest.argmin()]

    # sol = root(calc_v0_mBM5poly, vg, args=pars)
    # v0 = sol.x[0]
    sol = fsolve(calc_v0_mBM5poly, vg, args=pars)
    v0 = sol

    b0 = ((28 * e) / (9 * v0 ** (10 / 3)) +
          (2 * d) / v0 ** 3 + (10 * c) / (9 * v0 ** (8 / 3)) +
          (4 * b) / (9 * v0 ** (7 / 3))) * v0
    bp = (98 * e + 54 * d * v0 ** (1 / 3) + 25 * c * v0 ** (2 / 3) + 8 * b * v0) / \
        (42 * e + 27 * d * v0 ** (1 / 3) + 15 * c * v0 ** (2 / 3) + 6 * b * v0)
    b2p = (v0**(8 / 3) * (9 * d * (14 * e + 5 * c * v0**(2 / 3) + 8 * b * v0) +
                          2 * v0**(1 / 3) * (126 * b * e * v0 ** (1 / 3) + 5 * c * (28 * e + b * v0)))) / \
        (2 * (14 * e + 9 * d * v0**(1 / 3) + 5 * c * v0**(2 / 3) + 2 * b * v0)**3)

    '''
    dEdV = -b / 3 * v0**(-4 / 3) - 2 / 3 * c * \
        v0**(-5 / 3) - d * v0**(-2) - 4 / 3 * e * v0**(-7 / 3)
    dEdV2 = 4 / 9 * b * v0**(-7 / 3) + 10 / 9 * c * \
        v0**(-8 / 3) + 2 * d * v0**(-3) + 28 / 9 * e * v0**(-10 / 3)
    dEdV3 = -28 / 27 * b * \
        v0**(-10 / 3) - 80 / 27 * c * v0**(-11 / 3) - 6 * \
        d * v0**(-4) - 280 / 27 * e * v0**(-13 / 3)
    dEdV4 = 280 / 81 * b * v0**(-13 / 3) + 880 / 81 * c * \
        v0**(-14 / 3) + 24 * d * v0**(-5) + 3640 / 81 * e * v0**(-16 / 3)

    P = -v0 * dEdV
    dBdV = dEdV2 + v0 * dEdV3
    dPdV = -dEdV - v0 * dEdV2
    dBdV2 = dEdV3 + dEdV3 + v0 * dEdV4
    dPdV2 = -dEdV2 - dEdV2 - v0 * dEdV3

    b0 = v0 * dEdV2
    bp = dBdV / dPdV
    b2p = (dBdV2 * dPdV - dPdV2 * dBdV) / dPdV**3
    '''
    e0 = mBM5poly(v0, pars)
    props = [e0, b0, bp, v0, b2p]
    return props


# ----------------------------------------------------------------------------------------


def res_BM4(pars, y, x):
    return y - BM4(x, pars)


def calc_props_BM4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = (-143 + 63 * bp - 9 * bp * bp) / (9 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def BM4(vol, pars):
    """
    Birch-Murnaghan 4 pars equation from PRB 70, 224107, 3-order
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (v0 / vol) ** (1. / 3.)
    e = e0 + 9. * b0 * v0 / 16 * \
        (eta ** 2 - 1) ** 2 * (6 + bp * (eta ** 2 - 1.) - 4. * eta ** 2)
    return e


# ----------------------------------------------------------------------------------------
def res_BM5(pars, y, x):
    return y - BM5(x, pars)


def BM5(vol, pars):
    """
    Birch-Murnaghan 5 pars equation from PRB 70, 224107, 4-Order
    """
    e0 = pars[0]
    b0 = pars[1]
    b0p = pars[2]
    v0 = pars[3]
    b0pp = pars[4]

    t1 = (v0 / vol) ** (1. / 3.)
    t2 = t1 ** 2
    t3 = t2 - 1.
    t4 = t3 ** 2 / 4.
    t5 = b0p ** 2

    ee = e0 + 3. / 8. * b0 * v0 * t4 * (9. * t4 * b0 * b0pp +
                                        9. * t4 * t5 - 63. * t4 * b0p + 143. * t4 + 6. * b0p * t3 - 24. * t2 + 36.)

    return ee


def rBM4(vol, pars):
    '''
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = v0 / vol
    f = 0.5 * (x**(2. / 3) - 1)
    E = e0 + 4.5 * v0 * b0 * f**2 * (1 + (bp - 4) * f)
    return E


def res_rBM4(pars, y, x):
    res = y - rBM4(x, pars)
    return res


def rBM4_pv(vol, pars):
    '''
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    Fit for V-P relations
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = v0 / vol
    f = 0.5 * (x**(2. / 3) - 1)
    P = 1.5 * b0 * (2 * f + 1)**(2.5) * (2 + 3 * (bp - 4) * f)
    return P


def res_rBM4_pv(par, y, x):
    res = y - rBM4_pv(x, pars)
    return res


def rBM5(vol, pars):
    '''
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = v0 / vol
    f = 0.5 * (x**(2. / 3) - 1)
    H = b0 * bpp + b0 * b0

    E = e0 + 3 / 8 * v0 * b0 * f**2 * \
        ((9 * H - 63 * bp + 143) * f**2 + 12 * (bp - 4) * f + 12)
    return E


def res_rBM5(pars, y, x):
    res = y - rBM5(x, pars)
    return res


def rBM5_pv(vol, pars):
    '''
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    Fit for V-P relations
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = v0 / vol
    f = 0.5 * (x**(2. / 3) - 1)
    H = b0 * bpp + b0 * b0

    P = 0.5 * b0 * (2 * f + 1)**(2.5) * \
        ((9 * H - 63 * bp + 143) * f**2 + 9 * (bp - 4) * f + 6)
    return P


def res_rBM5_pv(par, y, x):
    res = y - rBM5_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------


def res_universal(pars, y, x):
    return y - universal(x, pars)


def universal(vol, parameters):
    """
    Universal equation of state(Vinet P et al., J. Phys.: Condens. Matter 1, p1941 (1989))
    """
    e0 = parameters[0]
    b0 = parameters[1]
    bp = parameters[2]
    v0 = parameters[3]

    t1 = b0 * v0
    t2 = bp - 1.
    t3 = (vol / v0) ** (1. / 3.)
    t4 = np.exp(-3. / 2. * t2 * (-1. + t3))
    t5 = t2 ** 2
    t6 = 1. / t5
    e = e0 - 2. * t1 * t4 * \
        (3. * t3 * bp - 3. * t3 + 5. - 3. * bp) * t6 + 4. * t1 * t6

    return e


# ----------------------------------------------------------------------------------------
def res_LOG4(pars, y, x):
    return y - LOG4(x, pars)


def LOG4(vol, pars):
    '''
    Natrual strain (Poirier-Tarantola)EOS with 4 paramters
    Seems only work in near-equillibrium range.
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    t1 = b0 * v0
    t2 = np.log(v0 / vol)
    t3 = t2 ** 2
    t4 = t3 * t2
    ee = e0 + t1 * t3 / 2. + t1 * t4 * bp / 6. - t1 * t4 / 3.
    '''
    # write follows ShunLi's
    xx = np.log(v0)
    a = e0 + b0 * v0 * (3 * xx**2 + (bp - 2) * xx**3) / 6
    b = -b0 * v0 * (2 * xx + (bp - 2) * xx**2) / 2
    c = b0 * v0 * (1 + (bp - 2) * xx) / 2
    d = -b0 * v0 * (bp - 2) / 6
    VV = np.log(vol)
    ee = a + b * VV + c * VV**2 + d * VV**3
    '''
    return ee


def calc_props_LOG4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (-3 + 3 * bp - bp * bp) / b0
    props = [e0, b0, bp, v0, bpp]
    return props

# ----------------------------------------------------------------------------------------


def rPT4(vol, pars):
    '''
    Natrual strain EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT4 for 4-parameters EOS.
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = vol / v0
    fn = 1 / 3 * np.log(x)
    E = e0 + 4.5 * b0 * v0 * fn**2 * (-(bp - 2) * fn + 1)
    return E


def res_rPT4(pars, y, x):
    res = y - rPT4(x, pars)
    return res


def rPT4_pv(vol, pars):
    '''
    Natrual strain (Poirier-Tarantola)EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT4 for 4-parameters EOS.
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = (vol / v0)
    fn = 1 / 3 * np.log(x)
    P = -1.5 * b0 * fn * np.exp(-3 * fn) * (-3 * (bp - 2) * fn + 1)
    return P


def res_rPT4_pv(pars, y, x):
    res = y - rPT4_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def res_LOG5(pars, y, x):
    return y - LOG5(x, pars)


def LOG5(vol, parameters):
    '''
    Natrual strain (Poirier-Tarantola)EOS with 5 paramters
    '''
    e0 = parameters[0]
    b0 = parameters[1]
    b0p = parameters[2]
    v0 = parameters[3]
    b0pp = parameters[4]

    t1 = b0 * v0
    t2 = np.log(v0 / vol)
    t3 = t2 ** 2
    t4 = t3 ** 2
    t5 = b0 ** 2
    t6 = b0p ** 2
    t7 = t3 * t2
    e = e0 + t1 * t4 / 8. + t5 * v0 * t4 * b0pp / 24. - t1 * t4 * b0p / 8. + \
        t1 * t4 * t6 / 24. + t1 * t7 * b0p / 6. - t1 * t7 / 3. + t1 * t3 / 2.

    return e


def rPT5(vol, pars):
    '''
    Natrual strain EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT5 for 4-parameters EOS.
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = (vol / v0)
    fn = 1 / 3 * np.log(x)
    H = b0 * bpp + bp * bp

    E = e0 + 9 / 8 * b0 * v0 * fn**2 * \
        (-3 * (H + 3 * bp - 3) * fn**2 - 4 * (bp - 2) * fn + 4)
    return E


def res_rPT5(pars, y, x):
    res = y - rPT5(x, pars)
    return res


def rPT5_pv(vol, pars):
    '''
    Natrual strain (Poirier-Tarantola)EOS with 5 paramters
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT5 for 4-parameters EOS.
    '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = (vol / v0)
    fn = 1 / 3 * np.log(x)
    H = b0 * bpp + bp * bp

    P = -1.5 * b0 * fn * \
        np.exp(-3 * fn) * (-3 * (H + 3 * bp - 3)
                           * fn**2 - 3 * (bp - 2) * fn + 2)
    return P


def res_rPT5_pv(pars, y, x):
    res = y - rPT5_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def res_vinet(pars, y, x):
    res = y - vinet(x, pars)
    return res


def calc_props_vinet(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (19 - 18 * bp - 9 * bp * bp) / (36 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def vinet(vol, pars):
    """
    Vinet equation from PRB 70, 224107
    Following, Shang Shunli et al., comput mater sci, 2010: 1040-1048, original expressions.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (vol / v0) ** (1 / 3)
    Y = 1.5 * (bp - 1) * (1 - eta)
    Z = 4 * b0 * v0 / (bp - 1)**2

    ee = e0 + Z - Z * (1 - Y) * np.exp(Y)
    return ee


def vinet_pv(vol, pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (vol / v0) ** (1 / 3)
    P = 3 * b0 * (1 - eta / eta**2) * np.exp(-1.5 * (bp - 1) * (eta - 1))
    return P


def res_vinet_pv(par, y, x):
    res = y - vinet(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def Li4p(V, parameters):
    ''' Li JH, APL, 87, 194111 (2005) '''
    E0 = parameters[0]
    B0 = parameters[1]
    BP = parameters[2]
    V0 = parameters[3]

    # In fact, it is just a small modified version of Rose or Vinet EOS.
    x = (V / V0) ** (1. / 3.)
    eta = np.sqrt(-9 * B0 * V0 / E0)
    astar = eta * (x - 1.)
    delta = (BP - 1.) / (2 * eta) - 1. / 3.

    E = E0 * (1 + astar + delta * astar ** 3) * np.exp(-astar)

    return E


def res_Li4p(p, y, x):
    return y - Li4p(x, p)


# ----------------------------------------------------------------------------------------
def morse(v, pars):
    ''' Reproduce from ShunliShang's matlab script. '''
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    # usually used in high pressure range, but maybe failed in large
    # expansion( > 2*v0) range.
    a = e0 + (9 * b0 * v0) / (2 * (-1 + bp) * (-1 + bp))
    b = (-9 * b0 * np.exp(-1 + bp) * v0) / (-1 + bp) / (-1 + bp)
    c = (9 * b0 * np.exp(-2 + 2 * bp) * v0) / (2 * (-1 + bp) * (-1 + bp))
    d = (1 - bp) / (v0 ** (1. / 3))

    ee = a + b * np.exp(d * np.power(v, 1. / 3)) + \
        c * np.exp(2 * d * np.power(v, 1. / 3))
    return ee


def calc_props_morse(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = (5 - 5 * bp - 2 * bp * bp) / (9 * b0)

    props = [e0, b0, bp, v0, bpp]
    return props


def res_morse(p, en, volume):
    res = en - morse(volume, p)
    return res


def morse_AB(volume, p):
    '''
    morse_AB EOS formula from Song's FVT souces
    '''
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    A = p[1]
    B = p[2]
    V0 = p[3]
    xx = (volume / V0) ** (1 / 3)
    E = 1.5 * V0 * A * (np.exp(B * (1. - 2.0 * xx)) -
                        2.0 * np.exp(-B * xx) + np.exp(-B)) + E0
    return E


def res_morse_AB(p, en, volume):
    res = en - morse_AB(volume, p)
    return res


def morse_3p(volume, p):
    '''
    morse_AB EOS formula from Song's FVT souces
    A= 0.5*B
    '''
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    A = p[1]
    V0 = p[2]
    B = 0.5 * A
    xx = (volume / V0) ** (1 / 3)
    E = 1.5 * V0 * A * (np.exp(B * (1. - 2.0 * xx)) -
                        2.0 * np.exp(-B * xx) + np.exp(-B)) + E0
    return E


def res_morse_3p(p, en, volume):
    res = en - morse_3p(volume, p)
    return res


def morse_6p(vol, par):
    '''
    Generalized Morse EOS proposed by Qin, see:
    Qin et al. Phys Rev B, 2008, 78, 214108.
    Qin et al. Phys Rev B, 2008, 77, 220103(R).
    '''
    # p0 = [e0, b0, bp, v0, bpp]
    e0 = par[0]
    b0 = par[1]
    p = abs(par[2])
    v0 = par[3]
    q = abs(par[4])
    m = abs(par[5])
    n = abs(par[6])

    C = (p + m) * (q - n) * ((p + m) - (q - n)) + (p * n + q * m)
    A = 9 * b0 * v0 * (q - n) / C
    B = 9 * b0 * v0 * (p + m) / C

    x = (vol / v0) ** (1 / 3)
    ee = A * np.exp(-p * (x - 1)) / (x ** m) - \
        B * (x ** n) * np.exp(-q * (x - 1)) + e0 - (A - B)
    return ee


def calc_props_morse_6p(par):
    e0 = par[0]
    b0 = par[1]
    p = abs(par[2])
    v0 = par[3]
    q = abs(par[4])
    m = abs(par[5])
    n = abs(par[6])

    bp = (p + q) / 3 + 1
    bpp = ((7 - p * q) / 9 - bp) / b0
    props = [e0, b0, bp, v0, bpp, m, n]
    return props


def res_morse_6p(p, en, volume):
    res = en - morse_6p(volume, p)
    return res


# ----------------------------------------------------------------------------------------
def mie(v, p):
    '''
    Mie model for song's FVT
    '''
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    m = p[1]
    n = p[2]
    V0 = p[3]

    xx = (V0 / v) ** (1 / 3)
    E = E0 / (n - m) * (n * (xx ** m) - m * (xx ** n))

    return E


def res_mie(p, e, v):
    res = e - mie(v, p)
    return res


def mie_simple(v, p):
    '''
    Mie_simple model for song's FVT
    '''
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    m = 4
    n = p[2]
    V0 = p[3]

    xx = (V0 / v) ** (1 / 3)
    E = E0 / (n - m) * (n * (xx ** m) - m * (xx ** n))

    return E


def res_mie_simple(p, e, v):
    res = e - mie_simple(v, p)
    return res


# ----------------------------------------------------------------------------------------
def TEOS(v, par):
    '''
    Holland, et al, Journal of Metamorphic Geology, 2011, 29(3): 333-383
    Modified Tait equation of Huang & Chow
    '''
    e0 = par[0]
    b0 = par[1]
    bp = par[2]
    v0 = par[3]
    bpp = par[4]

    a = (1 + bp) / (1 + bp + b0 * bpp)
    b = bp / b0 - bpp / (1 + bp)
    c = (1 + bp + b0 * bpp) / (bp ** 2 + bp - b0 * bpp)

    t1 = (v / v0) + a - 1
    P = 1 / b * ((t1 / a) ** (-1 / c) - 1)
    # ?? the following energy relation is nonsenses.
    E = -P * np.log(v0) + e0
    return E


def res_TEOS(p, e, v):
    res = e - TEOS(v, p)
    return res


# ----------------------------------------------------------------------------------------
def SJX_v2(vol, par):
    '''
    Sun Jiuxun, et al. J phys Chem Solids, 2005, 66: 773-782.
    They said it is satified for the limiting condition at high pressure.
    '''
    e0 = par[0]
    b0 = par[1]
    bp = par[2]
    v0 = par[3]

    Y = (v0 / vol)**(1 / 3) - 1
    alpha = 1 / 4 * (3 * bp - 1)
    beta = 1 - 1 / alpha
    a = alpha
    b = beta
    ee = e0 + 3 / 10 * a**4 * b0 * v0 * (2 * (b**5 + 1 / a**5) * (1 - (Y + 1)**(-3))
                                         - 15 * b**4 * (1 - (Y + 1)**(-2))
                                         - 60 * b**3 * (1 - (Y + 1)**(-1))
                                         + 60 * b**2 * np.log(Y + 1) + 3 * Y * (Y + 2) - 30 * b * Y)
    return ee


def res_SJX_v2(p, e, v):
    res = e - SJX_v2(v, p)
    return res


def SJX_5p(vol, par):
    '''
    SJX_5p's five parameters EOS, Physica B: Condens Mater, 2011, 406: 1276-1282
    '''
    e0 = par[0]
    a = par[1]
    b = par[2]
    v0 = par[3]
    n = par[4]

    X = (vol / v0) ** (1 / 3)
    C1 = n * b * np.exp(a * (1 - X) + b * (1 - X ** n))
    C2 = (-n * b - a) * np.exp(b * (1 - X ** n))
    ee = e0 / a * (C1 + C2)
    return ee


def res_SJX_5p(p, e, v):
    res = e - SJX_5p(v, p)
    return res


def calc_props_SJX_5p(par):
    e0 = par[0]
    a = par[1]
    b = par[2]
    v0 = par[3]
    n = par[4]

    b0 = n * b * e0 / (9 * v0) * (a + n * b + n - 1)
    Q = (a + n * b) * (a + 2 * n * b) - (n - 1) * (n - 2)
    bp = 1 + (Q / 3) / (a + n * b + n - 1)

    props = [e0, b0, bp, v0, n]
    return props

# ----------------------------------------------------------------------------------------
# some utility functions


def read_ve(fin):
    if not os.path.exists(fin):
        print('Could not find input file: [%s]' % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    vol = []
    en = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # now, we can read ve.dat & vec.dat
            # read the first two numbers.
            v = float(tmp[0])
            e = float(tmp[1])
            vol.append(v)
            en.append(e)
        else:
            pass
    return [vol, en]


def read_vlp(fin, fstart, fend):
    if not os.path.exists(fin):
        print('>> Could not find input file: [%s]' % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    print("\n** Totally, there are %d data points reading..." % nline)
    vol = []
    cella = []
    cellb = []
    cellc = []
    cellba = []
    cellca = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # read the DATA information
            v = float(tmp[0])
            if len(tmp) == 6:
                # for vlp.dat
                a = float(tmp[1])
                b = float(tmp[2])
                c = float(tmp[3])
                ba = float(tmp[4])
                ca = float(tmp[5])
            elif len(tmp) == 7:
                # for velp.dat
                e = float(tmp[1])
                a = float(tmp[2])
                b = float(tmp[3])
                c = float(tmp[4])
                ba = float(tmp[5])
                ca = float(tmp[6])
            vol.append(v)
            cella.append(a)
            cellb.append(b)
            cellc.append(c)
            cellba.append(ba)
            cellca.append(ca)
    print("\n** Vmin = %f, Vmax = %f" % (min(vol), max(vol)))

    # some special conditions
    if (fstart <= 0):
        print('\n** Data range input parameters must be positive values!')
        os.sys.exit(-1)
    if fstart > fend:
        if fend != -1:
            tmp = fstart
            fstart = fend
            fend = tmp
    if (fend > nline):
        print(
            '\n** EoSfit fit range exceed available data numbers, Reset it to be %d now.' % nline)
        fend = nline
    if (fstart > nline):
        print('EoSfit fit range exceed available data numbers, Reset it to be 1: %d now.' % nline)
        fstart = 1
        fend = -1

    fstart = fstart - 1

    # get data in the predefined range.
    fstart = fstart - 1
    if fend != -1:
        vol = vol[fstart:fend]
        cella = cella[fstart:fend]
        cellb = cellb[fstart:fend]
        cellc = cellc[fstart:fend]
        cellba = cellba[fstart:fend]
        cellca = cellca[fstart:fend]
    else:
        vol = vol[fstart:]
        cella = cella[fstart:]
        cellb = cellb[fstart:]
        cellc = cellc[fstart:]
        cellba = cellba[fstart:]
        cellca = cellca[fstart:]
    return [vol, cella, cellb, cellc, cellba, cellca]


def read_velp(fin, fstart, fend):
    if not os.path.exists(fin):
        print('>> Could not find input file: [%s]' % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    print("\n** Totally, there are %d data points reading..." % nline)
    vol = []
    eng = []
    cella = []
    cellb = []
    cellc = []
    cellba = []
    cellca = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # read the DATA information
            v = float(tmp[0])
            if len(tmp) == 7:
                # for velp.dat
                e = float(tmp[1])
                a = float(tmp[2])
                b = float(tmp[3])
                c = float(tmp[4])
                ba = float(tmp[5])
                ca = float(tmp[6])
            vol.append(v)
            eng.append(e)
            cella.append(a)
            cellb.append(b)
            cellc.append(c)
            cellba.append(ba)
            cellca.append(ca)
    print("\n** Vmin = %f, Vmax = %f" % (min(vol), max(vol)))

    # some special conditions
    if (fstart <= 0):
        print('\n** Data range input parameters must be positive values!')
        os.sys.exit(-1)
    if fstart > fend:
        if fend != -1:
            tmp = fstart
            fstart = fend
            fend = tmp
    if (fend > nline):
        print(
            '\n** EoSfit fit range exceed available data numbers, Reset it to be %d now.' % nline)
        fend = -1
    if (fstart > nline):
        print('EoSfit fit range exceed available data numbers, Reset it to be 1: %d now.' % nline)
        fstart = 1
        fend = -1

    # get data in the predefined range.
    fstart = fstart - 1
    if fend != -1:
        vol = vol[fstart:fend]
        eng = eng[fstart:fend]
        cella = cella[fstart:fend]
        cellb = cellb[fstart:fend]
        cellc = cellc[fstart:fend]
        cellba = cellba[fstart:fend]
        cellca = cellca[fstart:fend]
    else:
        vol = vol[fstart:]
        eng = eng[fstart:]
        cella = cella[fstart:]
        cellb = cellb[fstart:]
        cellc = cellc[fstart:]
        cellba = cellba[fstart:]
        cellca = cellca[fstart:]

    return [vol, eng, cella, cellb, cellc, cellba, cellca]


def init_guess(fin):
    v, e = read_ve(fin)
    a, b, c = np.polyfit(v, e, 2)  # this comes from pylab
    # initial guesses.
    v0 = np.abs(-b / (2 * a))
    e0 = a * v0 ** 2 + b * v0 + c
    b0 = 2 * a * v0
    bp = 3.0
    bpp = 1 * eV2GPa
    x0 = [e0, b0, bp, v0, bpp]
    return x0


def repro_ve(func, vol_i, p):
    ndata = len(vol_i)
    eni = np.zeros(ndata)
    for i in range(ndata):
        eni[i] = eval(func)(vol_i[i], p)
    return eni


def repro_vp(func, vol_i, pars):
    dv = 1e-5
    efunc = eval(func)
    ndata = len(vol_i)
    Press = np.zeros(ndata)
    for i in range(ndata):
        v = vol_i[i]
        P = (efunc(v + 0.5 * dv, pars) - efunc(v - 0.5 * dv, pars)) / dv
        Press[i] = -P * eV2GPa
    return Press


def ext_vec(func, fin, p0, fs, fe, vols=None, vole=None, ndata=101, refit=0, show_fig=False):
    '''
    extrapolate the data points for E-V based on the fitted parameters in small or
    very large volume range.
    '''
    # read fitted-parameters
    # pars = np.loadtxt(fin, dtype=float)
    fout = 'EoSfit.out'
    pars = lsqfit_eos(func, fin, p0, fs, fe, fout, refit=refit)

    vol, eng, cella, cellb, cellc, cellba, cellca = read_velp(fin, fs, fe)
    sca = ext_splint(vol, cellca)

    # define extrapolate range
    print("\n** Vext_start = %f, Vext_end = %f, N_ext = %d" %
          (vols, vole, ndata))
    vol_ext = np.linspace(vols, vole, ndata)
    en_ext = eval(func)(vol_ext, pars)

    # en_ext = np.zeros(ndata)
    fout = 'ext_ve_' + func + '.dat'
    fw = open(fout, 'w+')
    fw.write("%d\n" % ndata)
    for i in range(ndata):
        vx = vol_ext[i]
        ex = en_ext[i]
        if vx <= min(vol):
            cax = cellca[0]
        elif vx >= max(vol):
            cax = cellca[-1]
        else:
            cax = sca(vx)
        fw.write("%f\t%f\t%f\n" % (vx, ex, cax))
        fw.flush()
    fw.close()

    # plot the results
    vol, en = read_ve(fin)
    plt.plot(vol, en, 'o-', vol_ext, en_ext, 'rd')
    plt.legend(['dft_calc', func + '_ext'], loc='best')
    plt.savefig('ext_ve_' + func + '.png')
    if show_fig:
    	plt.show()
    plt.close()
    print("\n>> Storing the extrapolate results in %s\n" % fout)
    print("\n>> DONE!")

    return


def ext_splint(xp, yp, order=3, method='unispl'):
    if method == 'interp1d':
        # old-implement of 1-D interpolation
        # could not used in extrapolate
        SPLINT = interp1d
        return SPLINT(xp, yp, order, bounds_error=False)
    elif method == 'piecepoly':
        SPLINT = PiecewisePolynomial
        return SPLINT(xp, yp, order)
    else:
        if method == 'unispl':
            # 1-D smoothing spline fit to a given set of data
            SPLINT = UnivariateSpline
        else:
            SPLINT = LSQUnivariateSpline
        return SPLINT(xp, yp, k=order)


def ext_velp(fin, fstart, fend, vols, vole, ndata, order=3, method='unispl', fout='ext_velp.dat', show_fig=False):
    '''
    extrapolate the lattice parameters based on input data
    '''
    # read file
    vol, eng, cella, cellb, cellc, cellba, cellca = read_velp(
        fin, fstart, fend)

    # define extrapolate range
    print("\n** Vext_start = %f, Vext_end = %f, N_ext = %d" %
          (vols, vole, ndata))
    vv = np.linspace(vols, vole, ndata)

    # spline order = 3 by default
    se = ext_splint(vol, eng, order, method)
    ee = se(vv)
    sa = ext_splint(vol, cella, order, method)
    cellaa = sa(vv)
    sb = ext_splint(vol, cellb, order, method)
    cellbb = sb(vv)
    sc = ext_splint(vol, cellc, order, method)
    cellcc = sc(vv)
    sba = ext_splint(vol, cellba, order, method)
    cellbaba = sba(vv)
    sca = ext_splint(vol, cellca, order, method)
    cellcaca = sca(vv)
    cellca_cal = cellcc / cellaa

    # plot the extrapolate results
    nfigure = 6
    lp_ext = [ee, cellaa, cellbb, cellcc, cellbaba, cellcaca, cellca_cal]
    lp_ori = [eng, cella, cellb, cellc, cellba,
              cellca, np.array(cellc) / np.array(cella)]
    lp_ylabel = ['E(eV)', 'a(A)', 'b(A)', 'c(A)', 'b/a', 'c/a', 'c_ext/a_ext']
    for i in range(nfigure):
        plt.subplot(nfigure, 1, i + 1)
        plt.ylabel(lp_ylabel[i])
        plt.plot(vol, lp_ori[i], 'o-', vv, lp_ext[i], 'rd')
        if i == 0:
            plt.legend(['ori', 'ext'], loc='best')
    plt.xlabel('Volume (A**3)')
    plt.savefig('ext_velp.png')
    if show_fig:
    	plt.show()
    plt.close()

    # save the fit data
    # get vba.dat
    fw = open(fout, 'w+')
    fw.write('#%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s\n' %
             ('volume', 'eng', 'cella', 'cellb', 'cellc', 'b/a', 'c/a', 'cext/aext'))
    for i in range(ndata):
        fw.write('%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\n' %
                 (vv[i], ee[i], cellaa[i], cellbb[i], cellcc[i], cellbaba[i], cellcaca[i], cellca_cal[i]))
        fw.flush()
    fw.close()
    print("\n>> Storing the extrapolate results in %s\n" % fout)
    print("\n>> DONE!")
    return


def lsqfit_eos(func, fin, par, fstart, fend, show_fig=False, fout='EoSfit.out', refit=-1):
    # make the screen output better.
    print('\n')
    print("\t>> We are using [ %s ] to fit the V-E relationship << \t" % func)
    print('\n')

    fs = fstart
    fe = fend
    p = par

    if fs < 0:
        print('start fitting range index must be a positive integer!')
        os.sys.exit(-1)

    # p0 = [e0, b0, bp, v0, bpp]
    if refit == -1:
        if (func == 'morse_AB') or (func == 'morse_3p'):
            A = 6
            B = 0.5 * A
            if func == 'morse_AB':
                p0 = [p[0], A, B, p[3]]
            else:
                p0 = [p[0], A, p[3]]
        elif func in ['mie', 'mie_simple']:
            p0 = [p[0], 4, 6, p[3]]
        elif func == 'morse_6p':
            P = 1
            Q = 2
            m = 1
            n = 1
            p0 = [p[0], p[1], P, p[3], Q, m, n]
        elif func == 'SJX_5p':
            alpha = 1
            beta = 1
            n = 1
            p0 = [p[0], alpha, beta, p[3], n]
        else:
            p0 = p
        print(
            '>> use initial guess of fitted-paramters by polynomial fitting results:\n')
    else:
        p0 = np.loadtxt(fout)
        print(
            '>> use initial guess of fitted-paramters by last fitting results:\n')
    print(p0)

    vol, en = read_ve(fin)
    ndata = len(vol)
    # To determine the data points included in the fitting
    if fe > ndata:
        fe = ndata
        print(
            '\n[WARNING]: f_end exceed available data numbers, Reset it to be %d now.' % ndata)
    if fs > ndata:
        print(
            '\n[WARNING]: f_start exceed available data numbers, Reset it to be 1: %d now.' % ndata)
        print('and Reset f_end to be %d now.' % ndata)
        fs = 1
        fe = ndata

    if fe == -1:
        print('\n[ATTENTIONS]: fend = -1')
        num = ndata - fs + 1
    else:
        num = fe - fs + 1

    vol = vol[fs - 1: fe]
    en = en[fs - 1: fe]
    print('\n%d/%d data was used in the fitting ...\n' % (num, ndata))

    #*************************************************************************
    # fit it: step 1.
    popt, pcov, infodict, mesg, ier = leastsq(
        eval('res_' + func), p0, args=(en, vol), full_output=1, maxfev=(len(p0) + 1) * 400)
    nfev = infodict['nfev']
    fvec = infodict['fvec']

    '''
    psi_i = sum(fvec**2)
    psi_min = min(fvec)
    pfactor = np.zeros(num)
    wi = (4 + 1) / num * psi_i / psi_min
    pfi = np.exp(-wi * wi)
    '''
    if ier not in [1, 2, 3, 4]:
        raise RuntimeError("Optimal parameters not found: " + mesg)
    #*************************************************************************

    print('*' * 80)
    print(">> fitted parameters (with %d iterations):" % nfev)
    # p0 = [e0, b0, bp, v0, bpp]
    if func == 'morse_AB':
        e0, A, B, v0 = popt
        print("%12s\t%12s\t%12s\t%12s" % ('V0(A**3)', 'A', 'B', 'E0(eV)'))
        print('%12f\t%12f\t%12f\t%12f\n' % (v0, A, B, e0))
    elif func == 'morse_3p':
        e0, A, v0 = popt
        B = 0.5 * A
        print("%12s\t%12s\t%12s\t%12s" % ('V0(A**3)', 'A', 'B', 'E0(eV)'))
        print('%12f\t%12f\t%12f\t%12f\n' % (v0, A, B, e0))
    elif func in ['mie', 'mie_simple']:
        e0, m, n, v0 = popt
        print("%12s\t%12s\t%12s\t%12s" % ('V0(A**3)', 'm', 'n', 'E0(eV)'))
        print('%12f\t%12f\t%12f\t%12f\n' % (v0, m, n, e0))
    elif func == 'morse_6p':
        e0, b0, bp, v0, bpp, m, n = calc_props_morse_6p(popt)
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print("%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s" %
              ('V0(A**3)', 'B0(GPa)', 'Bp', 'E0(eV)', 'Bpp(1/GPa)', 'm', 'n'))
        print('%12f\t%12f\t%12f\t%12f\t%12f\t%12f\t%12f\n' %
              (v0, b0, bp, e0, bpp, m, n))
    elif func == 'SJX_5p':
        e0, b0, bp, v0, n = calc_props_SJX_5p(popt)
        b0 = eV2GPa * b0
        print("%12s\t%12s\t%12s\t%12s\t%12s" %
              ('V0(A**3)', 'B0(GPa)', 'Bp', 'E0(eV)', 'n'))
        print('%12f\t%12f\t%12f\t%12f\t%12f\n' % (v0, b0, bp, e0, n))
    elif func in ['mBM4poly', 'mBM5poly', 'mBM4', 'LOG4', 'vinet', 'morse', 'BM4']:
        prop_func = eval('calc_props_' + func)
        e0, b0, bp, v0, bpp = prop_func(popt)
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print("%12s\t%12s\t%12s\t%12s\t%12s" %
              ('V0(A**3)', 'B0(GPa)', 'Bp', 'E0(eV)', 'Bpp(1/GPa)'))
        print('%12f\t%12f\t%12f\t%12f\t%12f\n' % (v0, b0, bp, e0, bpp))
    else:
        e0, b0, bp, v0, bpp = popt
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print("%12s\t%12s\t%12s\t%12s\t%12s" %
              ('V0(A**3)', 'B0(GPa)', 'Bp', 'E0(eV)', 'Bpp(1/GPa)'))
        print('%12f\t%12f\t%12f\t%12f\t%12f\n' % (v0, b0, bp, e0, bpp))

    # write the fitted results in fit.out
    fw = open(fout, 'w+')
    for i in range(len(popt)):
        fw.write("%f\n" % popt[i])
        fw.flush()
    fw.close()

    # data evaluate by fitted parameters --to be used in the plot
    vol_i = np.linspace(min(vol) * 0.95, max(vol) * 1.05, len(vol) * 2)
    en_i = repro_ve(func, vol_i, popt)

    print('*' * 80)
    # calculate fitted residuals and square errors
    res_opt = np.zeros(len(vol))
    for i in range(len(vol)):
        res_opt[i] = (fvec[i])**2
    fit_res = sum(res_opt)
    fit_var = np.var(fvec)
    fit_std = np.std(fvec)
    print("\nfitted residuals\t= %16e\n" % fit_res)
    print("fitted variations\t= %16e\n" % fit_var)
    print("standard deviations\t= %16e\n" % fit_std)
    # if fit_res > 1e-4:
    #    print("\n>> Residuals seems too large, please refit it by swithing argument --refit 1!\n")
    #    show = 'F'  # reset show tag, not to show the figure.
    plt.plot(vol, en, 'o', vol_i, en_i)
    plt.title('EoS fitted by: %s model' % str(func))
    plt.legend(['calc', func + '-fit'], loc='best')
    plt.xlabel('Volume (A**3)')
    plt.ylabel('Energy (eV)')
    plt.savefig('EoSfit_' + func + '.png')
    if show_fig:
        plt.show()
    print('*' * 80)
    plt.close()

    # reproduce data by the popt and write it into files.
    repro_en = repro_ve(func, vol, popt)
    repro_press = repro_vp(func, vol, popt)
    fve = open(func + '_ve_fit.dat', 'w+')
    fvp = open(func + '_vp_fit.dat', 'w+')
    fve.write('#%20s\t%20s\t%20s\t%20s\n' %
              ('volume(A**3)', 'energy(fit)', 'energy(cal)', 'dE(%)'))
    fvp.write('#%20s\t%20s\t%20s\t%20s\n'
              % ('volume(A**3)', 'pressure(GPa)', 'pressure(Mbar)', 'pressure(kbar)'))
    for i in range(len(vol)):
        fve.write('%20f\t%20f\t%20f\t%20f\n' %
                  (vol[i], repro_en[i], en[i], 100 * np.abs((en[i] - repro_en[i]) / en[i])))
        fve.flush()
        p_tmp = repro_press[i]
        fvp.write('%20f\t%20f\t%20f\t%20f\n' %
                  (vol[i], p_tmp, p_tmp / 100, p_tmp * 10))
        fvp.flush()
    fve.close()
    fvp.close()
    return popt


def parse_argument():
    parser = argparse.ArgumentParser(
        description=" Script to fit EOS in MTP calculations")
    parser.add_argument('choice', help='to run mfp, ext_vec, ext_velp?')
    parser.add_argument(
        'infile', help='input ve.dat, vec.dat, velp.dat, or vlp.dat')
    parser.add_argument('-eos', '--eos', default='morse_AB',
                        help='Please choose one of the EOS name in: ' + str(get_eos_list()))
    parser.add_argument(
        '--show', default='T', help='to show the fitted results[T] or not[F]?')
    parser.add_argument('-vr', '--vrange', type=float, nargs=3,
                        default=[30, 50, 101], help="extend range used for ext, e.g. [vols, vole, ndata]: 1 10 31")
    parser.add_argument(
        '-fr', '--frange', type=int, nargs=2, default=[1, 1000], help='data range to be fitted, e.g. [ns, ne]')
    parser.add_argument('--refit', type=int, default=-1,
                        help='fit the data with last fitted parameters as initial guess')
    parser.add_argument('-eord', '--extorder', type=int,
                        default=3, help='spline order used in extrapolate')
    parser.add_argument('-enum', '--extnum', type=int, default=20,
                        help='number of data to be extrapolate')
    parser.add_argument(
        '-v', '--version', action='version', version='%(prog)s ' + __version__())
    args = parser.parse_args()
    return(args)


# main
if __name__ == '__main__':
    args = parse_argument()
    cho = args.choice
    fin = args.infile
    func = args.eos
    SHOW = args.show
    REFIT = args.refit

    eos_list = get_eos_list()
    if func not in eos_list:
        print("Choose one of the following eos name:")
        print(eos_list)
        os.sys.exit(-1)
    else:
        pass
    # p0 = [e0, b0, bp, v0, bpp]

    # used in mfp choice
    fs = args.frange[0]
    fe = args.frange[1]

    if fs < 0:
        print("ERROR, start range index must be a positive integer!")
        os.sys.exit(-1)

    if cho == 'mfp':
        p0 = init_guess(fin)
        lsqfit_eos(
            func, fin, p0, fs, fe, SHOW, refit=REFIT)
    elif cho == 'ext_vec':
        # used in ext choice
        ExtVS = args.vrange[0]
        ExtVE = args.vrange[1]
        ExtNum = int(args.vrange[2])
        if ExtVS < 0 or ExtVE < 0 or ExtNum < 0:
            print("ERROR, range setting must be a positive value!")
            exit(1)
        # if ExtNum % 2 == 0:
        #    print("[WARNING]: ndata = %d, reset it to be %d" %
        #          (ExtNum, ExtNum + 1))
        #    ExtNum = ExtNum + 1
        p0 = init_guess(fin)
        ext_vec(func, fin, p0, fs, fe, ExtVS, ExtVE, ExtNum, refit=REFIT)
    elif cho == 'ext_velp':
        EORDER = args.extorder
        ExtVS = args.vrange[0]
        ExtVE = args.vrange[1]
        ExtNum = int(args.vrange[2])
        if ExtVS < 0 or ExtVE < 0 or ExtNum < 0:
            print("ERROR, range setting must be a positive value!")
            exit(1)
        if ExtNum % 2 == 0:
            print("[WARNING]: ndata = %d, reset it to be %d" %
                  (ExtNum, ExtNum + 1))
            ExtNum = ExtNum + 1
        ext_velp(fin, fs, fe, ExtVS, ExtVE, ExtNum,
                 order=EORDER, method='unispl')
    else:
        print('Unkown Choice, abort now ... ...')
        os.sys.exit(-1)



import numpy as np
import requests
import os,re
from dpgen import dlog
from dpgen.auto_test.lib import vasp
from dpgen.auto_test.lib import lammps
from dpgen.auto_test.lib.utils import cmd_append_log

lammps_task_type=['deepmd','meam','eam']

def voigt_to_stress(inpt) :
    ret = np.zeros((3,3))
    ret[0][0] = inpt[0]
    ret[1][1] = inpt[1]
    ret[2][2] = inpt[2]
    ret[0][1] = inpt[3]
    ret[0][2] = inpt[4]
    ret[1][2] = inpt[5]
    ret[2][0] = ret[0][2]
    ret[1][0] = ret[0][1]
    ret[2][1] = ret[1][2]
    return ret

def insert_data(task,task_type,username,file_name):
    assert task in ['eos','elastic','surf']
    assert task_type in ['vasp','deepmd']
    url='http://115.27.161.2:5000/insert_test_data?username=%s&expr_type=%s&data_type=%s' % (username,task_type,task)
    res = requests.post(url, data=open(file_name).read())
    print('Successful upload!')


def make_work_path(jdata,task,reprod_opt,static,user):

    task_type=jdata['task_type']
    conf_dir=jdata['conf_dir']
    conf_path = os.path.abspath(conf_dir)
    task_path = re.sub('confs', task, conf_path)

    if task_type=="vasp":
        if user:
            work_path=os.path.join(task_path, 'vasp-user_incar')
            assert(os.path.isdir(work_path))
            return work_path
        if static:
            if 'scf_incar' in jdata.keys():
                task_type=task_type+'-static-scf_incar'
            else:
                kspacing = jdata['vasp_params']['kspacing']
                task_type=task_type+'-static-k%.2f' % (kspacing)
        else:
            if 'relax_incar' in jdata.keys():
                task_type=task_type+'-relax_incar'
            else:
                kspacing = jdata['vasp_params']['kspacing']
                task_type=task_type+'-k%.2f' % (kspacing)
    elif task_type in lammps_task_type:
        if static:
            task_type=task_type+'-static'
        elif reprod_opt :
            if 'relax_incar' in jdata.keys():
                task_type=task_type+'-reprod-relax_incar'
            else:
                kspacing = jdata['vasp_params']['kspacing']
                task_type=task_type+'-reprod-k%.2f'% (kspacing)

    work_path=os.path.join(task_path, task_type)
    assert(os.path.isdir(work_path))
    return work_path


def get_machine_info(mdata,task_type):
    if task_type=="vasp":
        vasp_exec=mdata['fp_command']
        group_size = mdata['fp_group_size']
        resources = mdata['fp_resources']
        machine=mdata['fp_machine']
        command = vasp_exec
        command = cmd_append_log(command, "log")
    elif task_type in lammps_task_type:
        lmp_exec = mdata['lmp_command']
        group_size = mdata['model_devi_group_size']
        resources = mdata['model_devi_resources']
        machine=mdata['model_devi_machine']
        command = lmp_exec + " -i lammps.in"
        command = cmd_append_log(command, "model_devi.log")
    return machine, resources, command, group_size

def collect_task(all_task,task_type):

    if task_type == 'vasp':
        output_file ='OUTCAR'
        check_finished = vasp.check_finished
    elif task_type in lammps_task_type:
        output_file = 'log.lammps'
        check_finished = lammps.check_finished

    run_tasks_ = []
    for ii in all_task:
        fres = os.path.join(ii, output_file)
        if os.path.isfile(fres) :
            if not check_finished(fres):
                run_tasks_.append(ii)
        else :
            run_tasks_.append(ii)

    run_tasks = [os.path.basename(ii) for ii in run_tasks_]
    return run_tasks



#!/usr/bin/env python3

import random, os, sys, dpdata
import numpy as np
import subprocess as sp
import scipy.constants as pc
from distutils.version import LooseVersion

def _sample_sphere() :
    while True:
        vv = np.array([np.random.normal(), np.random.normal(), np.random.normal()])
        vn = np.linalg.norm(vv)
        if vn < 0.2 :
            continue
        return vv / vn

def make_lammps_input(ensemble, 
                      conf_file,
                      graphs,
                      nsteps,
                      dt,
                      neidelay,
                      trj_freq, 
                      mass_map,
                      temp, 
                      jdata,
                      tau_t = 0.1,
                      pres = None,
                      tau_p = 0.5,
                      pka_e = None,
                      ele_temp_f = None,
                      ele_temp_a = None,
                      max_seed = 1000000,
                      nopbc = False,
                      deepmd_version = '0.1') :
    if (ele_temp_f is not None or ele_temp_a is not None) and LooseVersion(deepmd_version) < LooseVersion('1'):
        raise RuntimeError('the electron temperature is only supported by deepmd-kit >= 1.0.0, please upgrade your deepmd-kit')
    if ele_temp_f is not None and ele_temp_a is not None:
        raise RuntimeError('the frame style ele_temp and atom style ele_temp should not be set at the same time')
    ret = "variable        NSTEPS          equal %d\n" % nsteps
    ret+= "variable        THERMO_FREQ     equal %d\n" % trj_freq
    ret+= "variable        DUMP_FREQ       equal %d\n" % trj_freq
    ret+= "variable        TEMP            equal %f\n" % temp
    if ele_temp_f is not None:
        ret+= "variable        ELE_TEMP        equal %f\n" % ele_temp_f
    if ele_temp_a is not None:
        ret+= "variable        ELE_TEMP        equal %f\n" % ele_temp_a
    ret+= "variable        PRES            equal %f\n" % pres
    ret+= "variable        TAU_T           equal %f\n" % tau_t
    ret+= "variable        TAU_P           equal %f\n" % tau_p
    ret+= "\n"
    ret+= "units           metal\n"
    if nopbc:
        ret+= "boundary        f f f\n"
    else:
        ret+= "boundary        p p p\n"
    ret+= "atom_style      atomic\n"
    ret+= "\n"
    ret+= "neighbor        1.0 bin\n"
    if neidelay is not None :
        ret+= "neigh_modify    delay %d\n" % neidelay
    ret+= "\n"
    ret+= "box          tilt large\n"
    ret+= "read_data       %s\n" % conf_file
    ret+= "change_box   all triclinic\n"
    for jj in range(len(mass_map)) :
        ret+= "mass            %d %f\n" %(jj+1, mass_map[jj])
    graph_list = ""
    for ii in graphs :
        graph_list += ii + " "
    if LooseVersion(deepmd_version) < LooseVersion('1'):
        # 0.x
        ret+= "pair_style      deepmd %s ${THERMO_FREQ} model_devi.out\n" % graph_list
    else:
        # 1.x
        keywords = ""
        if jdata.get('use_clusters', False):
            keywords += "atomic "
        if jdata.get('use_relative', False):
            eps = jdata.get('eps', 0.)
            keywords += "relative %s " % jdata['epsilon']
        if ele_temp_f is not None:
            keywords += "fparam ${ELE_TEMP}"
        if ele_temp_a is not None:
            keywords += "aparam ${ELE_TEMP}"
        ret+= "pair_style      deepmd %s out_freq ${THERMO_FREQ} out_file model_devi.out %s\n" % (graph_list, keywords)
    ret+= "pair_coeff      \n"
    ret+= "\n"
    ret+= "thermo_style    custom step temp pe ke etotal press vol lx ly lz xy xz yz\n"
    ret+= "thermo          ${THERMO_FREQ}\n"
    ret+= "dump            1 all custom ${DUMP_FREQ} traj/*.lammpstrj id type x y z\n"
    ret+= "\n"
    if pka_e is None :
        ret+= "velocity        all create ${TEMP} %d" % (random.randrange(max_seed-1)+1)
    else :
        sys = dpdata.System(conf_file, fmt = 'lammps/lmp')
        sys_data = sys.data
        pka_mass = mass_map[sys_data['atom_types'][0] - 1]
        pka_vn = pka_e * pc.electron_volt / \
                 (0.5 * pka_mass * 1e-3 / pc.Avogadro * (pc.angstrom / pc.pico) ** 2)
        pka_vn = np.sqrt(pka_vn)
        print(pka_vn)
        pka_vec = _sample_sphere()
        pka_vec *= pka_vn
        ret+= 'group           first id 1\n'
        ret+= 'velocity        first set %f %f %f\n' % (pka_vec[0], pka_vec[1], pka_vec[2])
        ret+= 'fix	       2 all momentum 1 linear 1 1 1\n'
    ret+= "\n"
    if ensemble.split('-')[0] == 'npt' :
        assert (pres is not None)
        if nopbc:
            raise RuntimeError('ensemble %s is conflicting with nopbc' % ensemble)
    if ensemble == "npt" or ensemble == "npt-i" or ensemble == "npt-iso" :
        ret+= "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n"
    elif ensemble == 'npt-a' or ensemble == 'npt-aniso' : 
        ret+= "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}\n"
    elif ensemble == 'npt-t' or ensemble == 'npt-tri' : 
        ret+= "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}\n"
    elif ensemble == "nvt" :
        ret+= "fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n"
    elif ensemble == 'nve' :
        ret+= "fix             1 all nve\n"
    else :
        raise RuntimeError("unknown emsemble " + ensemble)
    if nopbc:
        ret+= "velocity        all zero linear\n"
        ret+= "fix             fm all momentum 1 linear 1 1 1\n"
    ret+= "\n"
    ret+= "timestep        %f\n" % dt
    ret+= "run             ${NSTEPS}\n"
    return ret
        
# ret = make_lammps_input ("npt", "al.lmp", ['graph.000.pb', 'graph.001.pb'], 20000, 20, [27], 1000, pres = 1.0)
# print (ret)
# cvt_lammps_conf('POSCAR', 'tmp.lmp')


    
 



#!/usr/bin/env python3

import os,sys,json,glob,argparse
import numpy as np
import subprocess as sp

def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def collect_data(target_folder, param_file, output, verbose = True) :
    target_folder = os.path.abspath(target_folder)
    output = os.path.abspath(output)
    tool_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'template')    
    command_cvt_2_raw = os.path.join(tool_path, 'tools.vasp', 'convert2raw.py')
    command_cvt_2_raw += ' data.configs'
    command_shuffle_raw = os.path.join(tool_path, 'tools.raw', 'shuffle_raw.py')
    command_raw_2_set = os.path.join(tool_path, 'tools.raw', 'raw_to_set.sh')
    # goto input    
    cwd = os.getcwd()
    os.chdir(target_folder)
    jdata = json.load(open(param_file))
    sys = jdata['sys_configs']
    if verbose :
        max_str_len = max([len(str(ii)) for ii in sys])
        ptr_fmt = '%%%ds   %%6d' % (max_str_len+5)
    # collect systems from iter dirs
    coll_sys = [[] for ii in sys]    
    numb_sys = len(sys)
    iters = glob.glob('iter.[0-9]*[0-9]')
    iters.sort()    
    for ii in iters :
        iter_data = glob.glob(os.path.join(ii, '02.fp', 'data.[0-9]*[0-9]'))
        iter_data.sort()
        for jj in iter_data :
            sys_idx = int(os.path.basename(jj).split('.')[-1])
            coll_sys[sys_idx].append(jj)
    # create output dir
    os.makedirs(output, exist_ok = True)
    # loop over systems
    for idx,ii in enumerate(coll_sys) :
        if len(ii) == 0 :
            continue
        # link iter data dirs
        out_sys_path = os.path.join(output, 'system.%03d' % idx)
        os.makedirs(out_sys_path, exist_ok=True)
        cwd_ = os.getcwd()
        os.chdir(out_sys_path)
        for jj in ii :
            in_sys_path = os.path.join(target_folder, jj)
            in_iter = in_sys_path.split('/')[-3]
            in_base = in_sys_path.split('/')[-1]
            out_file = in_iter + '.' + in_base
            if os.path.exists(out_file) :
                os.remove(out_file)
            os.symlink(in_sys_path, out_file)
        # cat data.configs
        data_configs = glob.glob(os.path.join('iter.[0-9]*[0-9].data.[0-9]*[0-9]', 'orig', 'data.configs'))
        data_configs.sort()
        os.makedirs('orig', exist_ok = True)        
        with open(os.path.join('orig', 'data.configs'), 'w') as outfile:
            for fname in data_configs:
                with open(fname) as infile:
                    outfile.write(infile.read())        
        # convert to raw
        os.chdir('orig')
        sp.check_call(command_cvt_2_raw, shell = True)
        os.chdir('..')
        # shuffle raw
        sp.check_call(command_shuffle_raw + ' orig ' + ' . > /dev/null', shell = True)
        if os.path.exists('type.raw') :
            os.remove('type.raw')
        os.symlink(os.path.join('orig', 'type.raw'), 'type.raw')
        # raw to sets
        sp.check_call(command_raw_2_set + ' > /dev/null', shell = True)
        # print summary
        if verbose :
            ndata = file_len('box.raw')
            print(ptr_fmt % (str(sys[idx]), ndata))
        # ch dir
        os.chdir(cwd_)


def _main()   :
    parser = argparse.ArgumentParser(description='Collect data from DP-GEN iterations')
    parser.add_argument("JOB_DIR", type=str, 
                        help="the directory of the DP-GEN job")
    parser.add_argument("OUTPUT", type=str, 
                        help="the output directory of data")
    parser.add_argument('-p',"--parameter", type=str, default = 'param.json',
                        help="the json file provides DP-GEN paramters, should be located in JOB_DIR")
    parser.add_argument('-v',"--verbose", action = 'store_true',
                        help="print number of data in each system")
    args = parser.parse_args()
    
    collect_data(args.JOB_DIR, args.parameter, args.OUTPUT, args.verbose)        

if __name__ == '__main__':
    _main()

    



#!/usr/bin/env python3

import os, glob, argparse, json, re
import dpgen.auto_test.lib.util as util
import dpgen.auto_test.lib.lammps as lammps
import dpgen.auto_test.lib.vasp as vasp

global_task_name = '01.eos'

def comput_lmp_eos(jdata,conf_dir, task_name) :
    conf_path = re.sub('confs', global_task_name, conf_dir)
    conf_path = os.path.abspath(conf_path)
    conf_path = os.path.join(conf_path, task_name)
    vol_paths = glob.glob(os.path.join(conf_path, 'vol-*'))
    vol_paths.sort(key=lambda k : float(k.split('-')[-1]))
    result = os.path.join(conf_path,'result')
    print('Vpa(A^3)\tEpA(eV)')
    with open(result,'w') as fp:
        fp.write('conf_dir:%s\n VpA(A^3)  EpA(eV)\n'% (conf_dir))
        for ii in vol_paths :
            log_lammps = os.path.join(ii, 'log.lammps')
            natoms, epa, vpa = lammps.get_nev(log_lammps)
            print(vpa, epa)
            fp.write('%7.3f  %8.4f \n' % (vpa,epa))
    fp.close()
    if 'upload_username' in jdata.keys() and task_name =='deepmd':
        upload_username=jdata['upload_username']
        util.insert_data('eos','deepmd',upload_username,result)

def comput_vasp_eos(jdata, conf_dir) :
    conf_path = re.sub('confs', global_task_name, conf_dir)
    conf_path = os.path.abspath(conf_path)

    if 'relax_incar' in jdata.keys():
        vasp_str='vasp-relax_incar'
    else:
        kspacing = jdata['vasp_params']['kspacing']
        vasp_str='vasp-k%.2f' % kspacing
    task_path = os.path.join(conf_path, vasp_str)
    vol_paths = glob.glob(os.path.join(task_path, 'vol-*'))
    vol_paths.sort(key=lambda k : float(k.split('-')[-1]))
    result = os.path.join(task_path,'result')
    print('Vpa(A^3)\tEpA(eV)')
    with open(result,'w') as fp:
        fp.write('conf_dir:%s\n VpA(A^3)  EpA(eV)\n'% (conf_dir))
        for ii in vol_paths :
            outcar = os.path.join(ii, 'OUTCAR')
            natoms, epa, vpa = vasp.get_nev(outcar)
            print(vpa, epa)
            fp.write('%7.3f  %8.4f \n' % (vpa,epa))
    fp.close()
    if 'upload_username' in jdata.keys():
        upload_username=jdata['upload_username']
        util.insert_data('eos','vasp',upload_username,result)

def _main():
    parser = argparse.ArgumentParser(
        description="cmpt 01.eos")
    parser.add_argument('TASK', type=str,
                        choices = ['vasp', 'deepmd', 'meam'],
                        help='the task of generation, vasp or lammps')
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    parser.add_argument('CONF', type=str,
                        help='the path to conf')
    args = parser.parse_args()

    with open (args.PARAM, 'r') as fp :
        jdata = json.load (fp)

    if args.TASK == 'vasp':
        comput_vasp_eos(jdata, args.CONF)
    elif args.TASK == 'deepmd' :
        comput_lmp_eos(jdata,args.CONF, args.TASK)
    elif args.TASK == 'meam' :
        comput_lmp_eos(jdata,args.CONF, args.TASK)
    else :
        raise RuntimeError("unknow task ", args.TASK)

if __name__ == '__main__' :
    _main()



#%%
import os
import argparse
import json
from collections import defaultdict
from itertools import tee

class System(object):
    current_num_of_system = 0
    current_num_of_sub_systems = 0

    @property
    def index_system(self):
        return self._index_system

    @index_system.setter
    def index_system(self,value):
        self._index_system = value
    
    @classmethod
    def register_system(cls):
        cls.current_num_of_system+=1
        return cls.current_num_of_system-1
    
    @classmethod
    def register_sub_system(cls):
        cls.current_num_of_sub_systems+=1
        return cls.current_num_of_sub_systems-1

    def __init__(self, system_prefix=""):
        # print(files_list)
        # if sum(map_relations)>len(files_list):
        #     raise RuntimeError(
        #         "files_list not enough;sum(map_relations):%s>len(files_list):%s, %s" 
        #         % (sum(map_relations),len(files_list),files_list,))
        self.index_system = self.register_system()
        self.sub_system_list = []
        self.system_prefix = system_prefix
        self.current_idx2 = 0
    
    def add_sub_system(self,idx2, files_list):
        idx1 = self.register_sub_system()
        idx2 = self.current_idx2 
        self.sub_system_list.append((idx1, self.index_system, idx2, files_list))
        self.current_idx2 += 1
        
    def get_sub_system(self):
        return self.sub_system_list

    
class Iteration(object):
    current_num_of_itearation = 0
    current_num_of_sub_itearation = 0
        
    @property
    def index_iteration(self):
        return self._index_iteration # pylint: disable=no-member
    
    @index_iteration.setter
    def index_iteration(self, value):
        self._index_sub_iteration = value
    
    @classmethod
    def register_iteration(cls):
        cls.current_num_of_itearation+=1
        return cls.current_num_of_itearation-1
    
    @classmethod
    def register_sub_iteartion(cls):
        cls.current_num_of_sub_itearation +=1
        return cls.current_num_of_sub_itearation-1
        
    def __init__(self, 
                temps,
                nsteps_list=[500, 500, 1000, 1000, 3000, 3000, 6000, 6000],
                sub_iteration_num=8, 
                ensemble='npt', 
                press=[1.0, 10.0, 100.0, 1000.0, 5000.0, 10000.0, 20000.0, 50000.0],
                trj_freq=10):
        if len(nsteps_list) != sub_iteration_num:
            raise RuntimeError(f'{nsteps_list}, {sub_iteration_num}; length does not match')
        self.temps = temps
        self.index_iteration = self.register_iteration()
        self.nsteps_list=nsteps_list
        self.sub_iteration_num=sub_iteration_num
        self.ensemble=ensemble
        self.press = press
        self.trj_freq = trj_freq

    def gen_sub_iter(self, system_list):
        sub_iter_list = []
        for idx2 in range(self.sub_iteration_num):
            iter_dict = {}
            iter_dict['_idx'] = self.register_sub_iteartion()
            iter_dict['ensemble'] = self.ensemble
            iter_dict['nsteps'] = self.nsteps_list[idx2]
            iter_dict['press'] = self.press
            iter_dict['sys_idx'] = [ii[0] for ii in system_list if ii[2]==idx2]
            iter_dict['temps'] = self.temps
            iter_dict['trj_freq'] = self.trj_freq
            sub_iter_list.append(iter_dict)
        return sub_iter_list

def default_map_generator(map_list=[1,1,2,2,2,4,4,4], data_list=None):
    num = 0
    # if len(data_list) < sum(map_list):
    #     raise RuntimeError(f'{data_list} < {map_list};not enough structure to expore, data_list_too_short!')
    if (data_list is None) and ( all(el%10==0 for el in map_list) ):
        for ii in map_list:
            yield [f"{jj:0<5}?" for jj in range(num, num+ii//10)]
            num+=(ii//10)
    elif data_list:
        for ii in map_list:
            yield [data_list[jj] for jj in range(num, num+ii)]
            num += ii
    raise RuntimeError(f"{map_list} length is not enough")
    #   while True:
        # yield [data_list[jj] for jj in range(num, num+ii)]
        # num += ii 

def get_system_list(system_dict,
    map_list=[1,1,2,2,2,4,4,4],
    meta_iter_num=4, 
    sub_iteration_num=8, 
    map_iterator=None,
    file_name="POSCAR"):
    """
    :type map_iterator: Iterable use to generate sys_configs
    :Exmaple [['000000', '000001',], ['00000[2-9]',], ['00001?', '000020',],]
    """
    if sub_iteration_num != len(map_list):
        raise RuntimeError(f"{sub_iteration_num},{map_list};sub_iteration_num does not match the length of map_list")
    
    system_list = []
    for system_prefix,data_list in system_dict.items():
        if map_iterator is None:
            print('12', data_list)
            new_map_iterator = default_map_generator(map_list=map_list, data_list=data_list)
        else:
            origin_one, new_map_iterator = tee(map_iterator) # pylint: disable=unused-variable
        # tee means copy;new_map_generator will become a copy of map_iterator 
        system = System(system_prefix)
        for idx2 in range(sub_iteration_num):
            files_list = [os.path.join(system_prefix, jj) for jj in next(new_map_iterator)]
            system.add_sub_system(idx2=idx2, files_list=files_list)
        system_list.extend(system.get_sub_system())
    return system_list

def scan_files(scan_dir="./" ,file_name="POSCAR", min_allow_files_num=20):
    # will return
    # files_list=[]
    system_dict = defaultdict(list)
    for ii in os.walk(scan_dir):
        if file_name in ii[2]:
            system_prefix = os.path.dirname(ii[0])
            system_suffix = os.path.basename(ii[0])
            system_dict[system_prefix].append(os.path.join(system_suffix, file_name))
    for k,v in list(system_dict.items()):
        if len(v) < min_allow_files_num:
            del system_dict[k]
    return system_dict

# def gen_
    
def default_temps_generator(melt_point, temps_intervel=0.1, num_temps=5):
    temps = [50, ]
    last_temp = 0 
    for ii in range(num_temps-1): # pylint: disable=unused-variable
        last_temp = last_temp + temps_intervel*melt_point
        temps.append(last_temp)
    yield temps
    while True:
        temps = []
        for ii in range(num_temps):
            last_temp = last_temp + temps_intervel*melt_point
            temps.append(last_temp)
        yield temps

def get_model_devi_jobs(melt_point,
    system_list,
    nsteps_list=[500, 500, 1000, 1000, 3000, 3000, 6000, 6000],
    press=[1.0, 10.0, 100.0, 1000.0, 5000.0, 10000.0, 20000.0, 50000.0],
    meta_iter_num=4, 
    sub_iteration_num=8,
    temps_iterator=None,
    ensemble="npt",
    trj_freq=10,
    temps_intervel=0.1,
    num_temps=5):

    if temps_iterator is None:
        temps_iterator = default_temps_generator(melt_point=melt_point, 
            temps_intervel=temps_intervel, num_temps=num_temps)

    if len(nsteps_list) != sub_iteration_num:
        raise RuntimeError(f"{nsteps_list}, {sub_iteration_num};length do not match!")
    model_devi_jobs =[]
    for ii in range(meta_iter_num): # pylint: disable=unused-variable
        temps = next(temps_iterator)
        meta_iter = Iteration(temps=temps,
            nsteps_list=nsteps_list,
            sub_iteration_num=sub_iteration_num,
            ensemble=ensemble,
            press=press,
            trj_freq=trj_freq)
        model_devi_jobs.extend(meta_iter.gen_sub_iter(system_list))
    return model_devi_jobs

def get_sys_configs(system_list):
    sys_configs=[[] for ii in system_list]
    for t in system_list:
        sys_configs[t[0]]=t[3]
    return sys_configs

def get_init_data_sys(scan_dir='./', init_file_name='type.raw'):

    init_data_sys = []
    for t in os.walk(scan_dir):
        if init_file_name in t[2]:
            init_data_sys.append(t[0])
        else: 
            pass
    return init_data_sys


def get_basic_param_json(melt_point,
    out_param_filename='param_basic.json',
    scan_dir="./", 
    file_name='POSCAR',
    init_file_name='type.raw',
    min_allow_files_num=16,
    map_list=[1,1,2,2,2,4,4,4],
    meta_iter_num=4,
    sub_iteration_num=8,
    map_iterator=None,
    nsteps_list=[500, 500, 1000, 1000, 3000, 3000, 6000, 6000],
    press=[1.0, 10.0, 100.0, 1000.0, 5000.0, 10000.0, 20000.0, 50000.0],
    temps_iterator=None,
    ensemble="npt",
    trj_freq=10,
    temps_intervel=0.1,
    num_temps=5,):

    init_data_sys = get_init_data_sys(scan_dir=scan_dir, init_file_name=init_file_name)
    print(f"length of init_data_sys: {len(init_data_sys)} {init_data_sys}")
    system_dict = scan_files(scan_dir, file_name, min_allow_files_num)
    print(f"num of different systems: {len(system_dict)}")
    system_list =get_system_list(system_dict,
        map_list=map_list, 
        meta_iter_num=meta_iter_num, 
        sub_iteration_num=sub_iteration_num, 
        map_iterator=map_iterator,
        file_name=file_name)

    sys_configs = get_sys_configs(system_list)
    print(f"length of sys_configs: {len(sys_configs)}")
    model_devi_jobs = get_model_devi_jobs(melt_point=melt_point,
        system_list=system_list,
        nsteps_list=nsteps_list,
        press=press,
        meta_iter_num=meta_iter_num, 
        sub_iteration_num=sub_iteration_num, 
        temps_iterator=temps_iterator,
        ensemble=ensemble,
        trj_freq=trj_freq,
        temps_intervel=temps_intervel,
        num_temps=num_temps)
    param_dict={
        'init_data_sys': init_data_sys,
        'sys_configs':sys_configs,
        'model_devi_jobs':model_devi_jobs
    }
    with open(out_param_filename, 'w') as p:
        json.dump(param_dict, p, indent=4)

    return param_dict 
def _main():
    parser = argparse.ArgumentParser(description='Collect data from inputs and generate basic param.json')
    parser.add_argument("melt_point", type=float, help="melt_point")
    # parser.addparser.add_argument("JOB_DIR", type=str, help="the directory of the DP-GEN job")
    args = parser.parse_args()
    get_basic_param_json(melt_point=args.melt_point)
  
if __name__=='__main__':
    _main()

def auto_gen_param(args):
    if args.PARAM:
        with open(args.PARAM) as p:
            j = json.load(p)
        melt_point = j['melt_point']
        print('param_basic.json', get_basic_param_json(melt_point=melt_point))
    else:
        raise RuntimeError('must provide melt point or PARAM')

#%%
